# #############################################################################
# dataset.py
# =================
# Authors :
# Yohann PERRON [yohann.perron@gmail.com]
# Eric BEZZAM [ebezzam@gmail.com]
# #############################################################################

from hydra.utils import get_original_cwd
import numpy as np
import glob
import os
import torch
from abc import abstractmethod
from torch.utils.data import Dataset, Subset
from torchvision import datasets, transforms
from torchvision.transforms import functional as F
from lensless.hardware.trainable_mask import prep_trainable_mask, AdafruitLCD
from lensless.utils.simulation import FarFieldSimulator
from lensless.utils.io import load_image, load_psf, save_image
from lensless.utils.image import is_grayscale, resize, rgb2gray
import re
from lensless.hardware.utils import capture
from lensless.hardware.utils import display
from lensless.hardware.slm import set_programmable_mask, adafruit_sub2full
from datasets import load_dataset
from lensless.recon.rfft_convolve import RealFFTConvolve2D
from huggingface_hub import hf_hub_download
import cv2
from lensless.hardware.sensor import sensor_dict, SensorParam
from scipy.ndimage import rotate
import warnings
from waveprop.noise import add_shot_noise
from lensless.utils.image import shift_with_pad


def convert(text):
    return int(text) if text.isdigit() else text.lower()


def alphanum_key(key):
    return [convert(c) for c in re.split("([0-9]+)", key)]


def natural_sort(arr):
    return sorted(arr, key=alphanum_key)


# available datasets
available_datasets = {
    # -- DiffuserCam MirFlickr (7.58 GB) https://huggingface.co/datasets/bezzam/DiffuserCam-Lensless-Mirflickr-Dataset-NORM
    "diffusercam_mirflickr": {
        "size (GB)": 7.58,
        "huggingface_repo": "bezzam/DiffuserCam-Lensless-Mirflickr-Dataset-NORM",
        "psf": "psf.tiff",
        "single_channel_psf": True,
        "flipud": True,
        "flip_lensed": True,
        "downsample": 2,
        "downsample_lensed": 2,
    },
    # -- TapeCam MirFlickr (10.5 GB) https://huggingface.co/datasets/bezzam/TapeCam-Mirflickr-25K
    "tapecam_mirflickr": {
        "size (GB)": 10.5,
        "huggingface_repo": "bezzam/TapeCam-Mirflickr-25K",
        "psf": "psf.png",
        "display_res": [900, 1200],
        "alignment": {"top_left": [45, 95], "height": 250},
    },
    # -- DigiCam CelebA (33.9 GB) https://huggingface.co/datasets/bezzam/DigiCam-CelebA-26K
    "digicam_celeba": {
        "size (GB)": 33.9,
        "huggingface_repo": "bezzam/DigiCam-CelebA-26K",
        "psf": "psf_simulated.png",
        "rotate": True,
        "split_seed": 0,
        "downsample": 2,
        "alignment": {"crop": {"vertical": [0, 525], "horizontal": [265, 695]}},
        "simulation": {
            "scene2mask": 0.25,
            "mask2sensor": 0.002,
            "object_height": 0.33,
            "sensor": "rpi_hq",
            "snr_db": None,
            "downsample": None,
            "random_vflip": False,
            "random_hflip": False,
            "quantize": False,
            "vertical_shift": -117,
            "horizontal_shift": -25,
        },
    },
    # -- DigiCam MirFlickr (11.9 GB) https://huggingface.co/datasets/bezzam/DigiCam-Mirflickr-SingleMask-25K
    "digicam_mirflickr": {
        "size (GB)": 11.9,
        "huggingface_repo": "bezzam/DigiCam-Mirflickr-SingleMask-25K",
        "display_res": [900, 1200],
        "rotate": True,
        "alignment": {"top_left": [80, 100], "height": 200},
    },
    # DigiCam MirFlickr Mini (472 MB) https://huggingface.co/datasets/bezzam/DigiCam-Mirflickr-SingleMask-1K
    "digicam_mirflickr_mini": {
        "size (GB)": 0.472,
        "huggingface_repo": "bezzam/DigiCam-Mirflickr-SingleMask-1K",
        "display_res": [900, 1200],
        "rotate": True,
        "alignment": {"top_left": [80, 100], "height": 200},
    },
    # -- DigiCam MirFlickr Multimask (12 GB) https://huggingface.co/datasets/bezzam/DigiCam-Mirflickr-MultiMask-25K
    "digicam_mirflickr_multi": {
        "size (GB)": 12,
        "huggingface_repo": "bezzam/DigiCam-Mirflickr-MultiMask-25K",
        "display_res": [900, 1200],
        "rotate": True,
        "alignment": {"top_left": [80, 100], "height": 200},
    },
    # -- DigiCam MirFlickr Multimask Mini (477 MB) https://huggingface.co/datasets/bezzam/DigiCam-Mirflickr-MultiMask-1K
    "digicam_mirflickr_multi_mini": {
        "size (GB)": 0.477,
        "huggingface_repo": "bezzam/DigiCam-Mirflickr-MultiMask-25K",
        "display_res": [900, 1200],
        "rotate": True,
        "alignment": {"top_left": [80, 100], "height": 200},
    },
    # MultiLens MirFlickr Ambient (16.7 GB) https://huggingface.co/datasets/Lensless/MultiLens-Mirflickr-Ambient
    "multilens_mirflickr_ambient": {
        "size (GB)": 16.7,
        "huggingface_repo": "Lensless/MultiLens-Mirflickr-Ambient",
        "psf": "psf.png",
        "display_res": [600, 600],
        "alignment": {"top_left": [118, 220], "height": 123},
    },
    # MultiLens MirFlickr Ambient Mini (67.7 MB) https://huggingface.co/datasets/Lensless/MultiLens-Mirflickr-Ambient-100
    "multilens_mirflickr_ambient_mini": {
        "size (GB)": 0.0677,
        "huggingface_repo": "Lensless/MultiLens-Mirflickr-Ambient-100",
        "psf": "psf.png",
        "display_res": [600, 600],
        "alignment": {"top_left": [118, 220], "height": 123},
    },
    # Multilens MirFlickr Mini (427 MB) https://huggingface.co/datasets/Lensless/mirflickr_voronoi_1k
    "multilens_mirflickr_mini": {
        "size (GB)": 0.427,
        "huggingface_repo": "Lensless/mirflickr_voronoi_1k",
        "psf": "psf_measured.png",
        "display_res": [900, 1200],
    },
    # Coded Aperture (MLS) MirFlickr 1K (467 MB) https://huggingface.co/datasets/Lensless/mirflickr_CA_fine_1k
    "mls_mirflickr_1k": {
        "size (GB)": 0.467,
        "huggingface_repo": "Lensless/mirflickr_CA_fine_1k",
        "psf": "psf_measured.png",
        "display_res": [900, 1200],
        # "alignment": {"top_left": [118, 220], "height": 123},
    },
    # Fresnel Zone Aperture MirFlickr 1K (454 MB) https://huggingface.co/datasets/Lensless/Mirflickr_FZA_fine_1k
    "fza_mirflickr_1k": {
        "size (GB)": 0.454,
        "huggingface_repo": "Lensless/Mirflickr_FZA_fine_1k",
        "psf": "psf_measured.png",
        "display_res": [900, 1200],
        # "alignment": {"top_left": [118, 220], "height": 123},
    },
}


def print_available_datasets():
    print("Available datasets:")
    for dataset in available_datasets:
        print(
            f"  - {dataset} ({available_datasets[dataset]['size (GB)']} GB) : https://huggingface.co/datasets/{available_datasets[dataset]['huggingface_repo']}"
        )


def get_dataset(dataset_name, split, **kwargs):
    """
    Get a dataset by name.

    Parameters
    ----------
    dataset_name : str
        Name of the dataset from the available datasets in ``available_datasets``.
    split : str
        Split of the dataset to load (e.g. "train", "test").

    Returns
    -------
    :py:class:`~torch.utils.data.Dataset`
        Dataset object.
    """
    if dataset_name not in available_datasets:
        print_str_available_dataset = "Available datasets are:"
        for dataset in available_datasets:
            print_str_available_dataset += f"\n  - {dataset} ({available_datasets[dataset]['size (GB)']} GB) : https://huggingface.co/datasets/{available_datasets[dataset]['huggingface_repo']}"
        raise ValueError(
            f"Dataset '{dataset_name}' not available.\n\n{print_str_available_dataset}"
        )
    assert split in ["train", "test"], "Split should be 'train' or 'test'"

    dataset_config = available_datasets[dataset_name]
    # replace dataset_config with anything from kwargs
    dataset_config.update(kwargs)
    return HFDataset(split=split, **dataset_config)


class DualDataset(Dataset):
    """
    Abstract class for defining a dataset of paired lensed and lensless images.
    """

    def __init__(
        self,
        indices=None,
        # psf_path=None,
        background=None,
        # background_pix=(0, 15),
        downsample=1,
        flip=False,
        flip_ud=False,
        flip_lr=False,
        transform_lensless=None,
        transform_lensed=None,
        input_snr=None,
        **kwargs,
    ):
        """
        Dataset consisting of lensless and corresponding lensed image.

        Parameters
        ----------
        indices : range or int or None
            Indices of the images to use in the dataset (if integer, it should be interpreted as range(indices)), by default None.
        psf_path : str
            Path to the PSF of the imaging system, by default None.
        background : :py:class:`~torch.Tensor` or None, optional
            If not ``None``, background is removed from lensless images, by default ``None``. If PSF is provided, background is estimated from the PSF.
        background_pix : tuple, optional
            Pixels to use for background estimation, by default (0, 15).
        downsample : int, optional
            Downsample factor of the lensless images, by default 1.
        flip : bool, optional
            If ``True``, lensless images are flipped, by default ``False``.
        transform_lensless : PyTorch Transform or None, optional
            Transform to apply to the lensless images, by default ``None``. Note that this transform is applied on HWC images (different from torchvision).
        transform_lensed : PyTorch Transform or None, optional
            Transform to apply to the lensed images, by default ``None``. Note that this transform is applied on HWC images (different from torchvision).
        input_snr : float, optional
            If not ``None``, Poisson noise is added to the lensless images to match the given SNR.
        """
        if isinstance(indices, int):
            indices = range(indices)
        self.indices = indices
        self.background = background
        self.input_snr = input_snr
        self.downsample = downsample
        self.flip = flip
        self.flip_ud = flip_ud
        self.flip_lr = flip_lr
        self.transform_lensless = transform_lensless
        self.transform_lensed = transform_lensed

        # self.psf = None
        # if psf_path is not None:
        #     psf, background = load_psf(
        #         psf_path,
        #         downsample=downsample,
        #         return_float=True,
        #         return_bg=True,
        #         bg_pix=background_pix,
        #     )
        #     if self.background is None:
        #         self.background = background
        #     self.psf = torch.from_numpy(psf)
        #     if self.transform_lensless is not None:
        #         self.psf = self.transform_lensless(self.psf)

    @abstractmethod
    def __len__(self):
        """
        Abstract method to get the length of the dataset. It should take into account the indices parameter.
        """
        raise NotImplementedError

    @abstractmethod
    def _get_images_pair(self, idx):
        """
        Abstract method to get the lensed and lensless images. Should return a pair (lensless, lensed) of numpy arrays with values in [0,1].

        Parameters
        ----------
        idx : int
            images index
        """
        raise NotImplementedError

    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.item()

        if self.indices is not None:
            idx = self.indices[idx]
        lensless, lensed = self._get_images_pair(idx)

        if isinstance(lensless, np.ndarray):
            # expected case
            if self.downsample != 1.0:
                lensless = resize(lensless, factor=1 / self.downsample)
                lensed = resize(lensed, factor=1 / self.downsample)

            lensless = torch.from_numpy(lensless)
            lensed = torch.from_numpy(lensed)
        else:
            # torch tensor
            # This mean get_images_pair returned a torch tensor. This isn't recommended, if possible get_images_pair should return a numpy array
            # In this case it should also have applied the downsampling
            pass

        # If [H, W, C] -> [D, H, W, C]
        if len(lensless.shape) == 3:
            lensless = lensless.unsqueeze(0)
        if len(lensed.shape) == 3:
            lensed = lensed.unsqueeze(0)

        if self.background is not None:
            lensless = lensless - self.background
            lensless = torch.clamp(lensless, min=0)

        # add noise
        if self.input_snr is not None:
            lensless = add_shot_noise(lensless, self.input_snr)

        # flip image x and y if needed
        if self.flip:
            lensless = torch.rot90(lensless, dims=(-3, -2), k=2)
            lensed = torch.rot90(lensed, dims=(-3, -2), k=2)
        if self.flip_ud:
            lensless = torch.flip(lensless, dims=(-4, -3))
            lensed = torch.flip(lensed, dims=(-4, -3))
        if self.flip_lr:
            lensless = torch.flip(lensless, dims=(-4, -2))
            lensed = torch.flip(lensed, dims=(-4, -2))
        if self.transform_lensless:
            lensless = self.transform_lensless(lensless)
        if self.transform_lensed:
            lensed = self.transform_lensed(lensed)

        return lensless, lensed


class SimulatedFarFieldDataset(DualDataset):
    """
    Dataset of propagated images (through simulation) from a Torch Dataset. :py:class:`lensless.utils.simulation.FarFieldSimulator` is used for simulation,
    assuming a far-field propagation and a shift-invariant system with a single point spread function (PSF).

    """

    def __init__(
        self,
        dataset,
        simulator,
        pre_transform=None,
        dataset_is_CHW=False,
        flip=False,
        vertical_shift=None,
        horizontal_shift=None,
        crop=None,
        downsample=1,
        **kwargs,
    ):
        """
        Parameters
        ----------

        dataset : :py:class:`torch.utils.data.Dataset`
            Dataset to propagate. Should output images with shape [H, W, C] unless ``dataset_is_CHW`` is ``True`` (and therefore images have the dimension ordering of [C, H, W]).
        simulator : :py:class:`lensless.utils.simulation.FarFieldSimulator`
            Simulator object used on images from ``dataset``. Waveprop simulator to use for the simulation. It is expected to have ``is_torch = True``.
        pre_transform : PyTorch Transform or None, optional
            Transform to apply to the images before simulation, by default ``None``. Note that this transform is applied on HCW images (different from torchvision).
        dataset_is_CHW : bool, optional
            If True, the input dataset is expected to output images with shape [C, H, W], by default ``False``.
        flip : bool, optional
            If True, images are flipped beffore the simulation, by default ``False``.
        """

        # we do the flipping before the simualtion
        super(SimulatedFarFieldDataset, self).__init__(flip=False, **kwargs)

        assert isinstance(dataset, Dataset)
        self.dataset = dataset
        self.n_files = len(dataset)
        self.dataset_is_CHW = dataset_is_CHW
        self._pre_transform = pre_transform
        self.flip_pre_sim = flip

        self.vertical_shift = vertical_shift
        self.horizontal_shift = horizontal_shift
        self.crop = crop.copy() if crop is not None else None
        if downsample != 1:
            if self.vertical_shift is not None:
                self.vertical_shift = int(self.vertical_shift // downsample)
            if self.horizontal_shift is not None:
                self.horizontal_shift = int(self.horizontal_shift // downsample)

            if crop is not None:
                self.crop["vertical"][0] = int(self.crop["vertical"][0] // downsample)
                self.crop["vertical"][1] = int(self.crop["vertical"][1] // downsample)
                self.crop["horizontal"][0] = int(self.crop["horizontal"][0] // downsample)
                self.crop["horizontal"][1] = int(self.crop["horizontal"][1] // downsample)

        # check simulator
        assert isinstance(simulator, FarFieldSimulator), "Simulator should be a FarFieldSimulator"
        assert simulator.is_torch, "Simulator should be a pytorch simulator"
        assert simulator.fft_shape is not None, "Simulator should have a psf"
        self.sim = simulator

    @property
    def psf(self):
        return self.sim.get_psf()

    def get_image(self, index):
        return self.dataset[index]

    def _get_images_pair(self, index):
        # load image
        img, _ = self.get_image(index)
        # convert to HWC for simulator and transform
        if self.dataset_is_CHW:
            img = img.moveaxis(-3, -1)
        if self.flip_pre_sim:
            img = torch.rot90(img, dims=(-3, -2))
        if self._pre_transform is not None:
            img = self._pre_transform(img)

        lensless, lensed = self.sim.propagate_image(img, return_object_plane=True)

        if self.vertical_shift is not None:
            lensed = torch.roll(lensed, self.vertical_shift, dims=-3)
        if self.horizontal_shift is not None:
            lensed = torch.roll(lensed, self.horizontal_shift, dims=-2)

        if lensed.shape[-1] == 1 and lensless.shape[-1] == 3:
            # copy to 3 channels
            lensed = lensed.repeat(1, 1, 3)
        assert (
            lensed.shape[-1] == lensless.shape[-1]
        ), "Lensed and lensless should have same number of channels"

        return lensless, lensed

    def __len__(self):
        if self.indices is None:
            return self.n_files
        else:
            return len([x for x in self.indices if x < self.n_files])


class MeasuredDatasetSimulatedOriginal(DualDataset):
    """
    Abstract class for defining a dataset of paired lensed and lensless images.

    Dataset consisting of lensless image captured from a screen and the corresponding image shown on the screen.
    Unlike :py:class:`lensless.utils.dataset.MeasuredDataset`, the ground-truth lensed image is simulated using a :py:class:`lensless.utils.simulation.FarFieldSimulator`
    object rather than measured with a lensed camera.

    The class assumes that the ``measured_dir`` and ``original_dir`` have file names that match.

    The method ``_get_images_pair`` must be defined.
    """

    def __init__(
        self,
        measured_dir,
        original_dir,
        simulator,
        measurement_ext="png",
        original_ext="jpg",
        downsample=1,
        background=None,
        flip=False,
        **kwargs,
    ):
        """
        Dataset consisting of lensless image captured from a screen and the corresponding image shown on screen.

        Parameters
        ----------
        """
        super(MeasuredDatasetSimulatedOriginal, self).__init__(
            downsample=1, background=background, flip=flip, **kwargs
        )
        self.pre_downsample = downsample

        self.measured_dir = measured_dir
        self.original_dir = original_dir
        assert os.path.isdir(self.measured_dir)
        assert os.path.isdir(self.original_dir)

        self.measurement_ext = measurement_ext.lower()
        self.original_ext = original_ext.lower()

        files = natural_sort(glob.glob(os.path.join(self.measured_dir, "*." + measurement_ext)))

        self.files = [os.path.basename(fn) for fn in files]

        if len(self.files) == 0:
            raise FileNotFoundError(
                f"No files found in {self.measured_dir} with extension {self.measurement_ext}"
            )

        # check that corresponding files exist
        for fn in self.files:
            original_fp = os.path.join(self.original_dir, fn[:-3] + self.original_ext)
            assert os.path.exists(original_fp), f"File {original_fp} does not exist"

        # check simulator
        assert isinstance(simulator, FarFieldSimulator), "Simulator should be a FarFieldSimulator"
        assert simulator.is_torch, "Simulator should be a pytorch simulator"
        assert simulator.fft_shape is None, "Simulator should not have a psf"
        self.sim = simulator

    def __len__(self):
        if self.indices is None:
            return len(self.files)
        else:
            return len([i for i in self.indices if i < len(self.files)])

    # def _get_images_pair(self, idx):
    #     if self.image_ext == "npy" or self.image_ext == "npz":
    #         lensless_fp = os.path.join(self.lensless_dir, self.files[idx])
    #         original_fp = os.path.join(self.original_dir, self.files[idx])
    #         lensless = np.load(lensless_fp)
    #         lensless = resize(lensless, factor=1 / self.downsample)
    #         original = np.load(original_fp[:-3] + self.original_ext)
    #     else:
    #         # more standard image formats: png, jpg, tiff, etc.
    #         lensless_fp = os.path.join(self.lensless_dir, self.files[idx])
    #         original_fp = os.path.join(self.original_dir, self.files[idx])
    #         lensless = load_image(lensless_fp, downsample=self.pre_downsample)
    #         original = load_image(
    #             original_fp[:-3] + self.original_ext, downsample=self.pre_downsample
    #         )

    #         # convert to float
    #         if lensless.dtype == np.uint8:
    #             lensless = lensless.astype(np.float32) / 255
    #             original = original.astype(np.float32) / 255
    #         else:
    #             # 16 bit
    #             lensless = lensless.astype(np.float32) / 65535
    #             original = original.astype(np.float32) / 65535

    #     # convert to torch
    #     lensless = torch.from_numpy(lensless)
    #     original = torch.from_numpy(original)

    #     # project original image to lensed space
    #     with torch.no_grad():
    #         lensed = self.sim.propagate_image()

    #     return lensless, lensed


class DigiCamCelebA(MeasuredDatasetSimulatedOriginal):
    def __init__(
        self,
        celeba_root,
        data_dir=None,
        psf_path=None,
        downsample=1,
        flip=True,
        vertical_shift=None,
        horizontal_shift=None,
        crop=None,
        simulation_config=None,
        **kwargs,
    ):
        """

        Some parameters default to work for the ``celeba_adafruit_random_2mm_20230720_10K`` dataset,
        namely: flip, vertical_shift, horizontal_shift, crop, simulation_config.

        Parameters
        ----------
        celeba_root : str
            Path to the CelebA dataset.
        data_dir : str, optional
            Path to the lensless images, by default looks inside the ``data`` folder. Can download if not available.
        psf_path : str, optional
            Path to the PSF of the imaging system, by default looks inside the ``data/psf`` folder. Can download if not available.
        downsample : int, optional
            Downsample factor of the lensless images, by default 1.
        flip : bool, optional
            If True, measurements are flipped, by default ``True``. Does not get applied to the original images.
        vertical_shift : int, optional
            Vertical shift (in pixels) of the lensed images to align.
        horizontal_shift : int, optional
            Horizontal shift (in pixels) of the lensed images to align.
        crop : dict, optional
            Dictionary of crop parameters (vertical: [start, end], horizontal: [start, end]) to select region of interest.
        """

        if vertical_shift is None:
            # default to (no downsampling) of celeba_adafruit_random_2mm_20230720_10K
            vertical_shift = -85
            horizontal_shift = -5

        if crop is None:
            crop = {"vertical": [30, 560], "horizontal": [285, 720]}
        self.crop = crop

        self.vertical_shift = vertical_shift
        self.horizontal_shift = horizontal_shift
        if downsample != 1:
            self.vertical_shift = int(self.vertical_shift // downsample)
            self.horizontal_shift = int(self.horizontal_shift // downsample)

            self.crop["vertical"][0] = int(self.crop["vertical"][0] // downsample)
            self.crop["vertical"][1] = int(self.crop["vertical"][1] // downsample)
            self.crop["horizontal"][0] = int(self.crop["horizontal"][0] // downsample)
            self.crop["horizontal"][1] = int(self.crop["horizontal"][1] // downsample)

        # download dataset if necessary
        if data_dir is None:
            data_dir = os.path.join(
                os.path.dirname(__file__),
                "..",
                "..",
                "data",
                "celeba_adafruit_random_2mm_20230720_10K",
            )
        if not os.path.isdir(data_dir):
            main_dir = os.path.join(os.path.dirname(__file__), "..", "..", "data")
            print("DigiCam CelebA dataset not found.")
            try:
                from torchvision.datasets.utils import download_and_extract_archive
            except ImportError:
                exit()
            msg = "Do you want to download this dataset of 10K examples (12.2GB)?"

            # default to yes if no input is given
            valid = input("%s (Y/n) " % msg).lower() != "n"
            if valid:
                url = "https://drive.switch.ch/index.php/s/9NNGCJs3DoBDGlY/download"
                filename = "celeba_adafruit_random_2mm_20230720_10K.zip"
                download_and_extract_archive(url, main_dir, filename=filename, remove_finished=True)

        # download PSF if necessary
        if psf_path is None:
            psf_path = os.path.join(
                os.path.dirname(__file__),
                "..",
                "..",
                "data",
                "psf",
                "adafruit_random_2mm_20231907.png",
            )
        if not os.path.exists(psf_path):
            try:
                from torchvision.datasets.utils import download_url
            except ImportError:
                exit()
            msg = "Do you want to download the PSF (38.8MB)?"

            # default to yes if no input is given
            valid = input("%s (Y/n) " % msg).lower() != "n"
            output_path = os.path.join(os.path.dirname(__file__), "..", "..", "data", "psf")
            if valid:
                url = "https://drive.switch.ch/index.php/s/kfN5vOqvVkNyHmc/download"
                filename = "adafruit_random_2mm_20231907.png"
                download_url(url, output_path, filename=filename)

        # load PSF
        self.flip_measurement = flip
        psf, background = load_psf(
            psf_path,
            downsample=downsample * 4,  # PSF is 4x the resolution of the images
            return_float=True,
            return_bg=True,
            flip=flip,
            bg_pix=(0, 15),
        )
        self.psf = torch.from_numpy(psf)

        # create simulator
        simulation_config["output_dim"] = tuple(self.psf.shape[-3:-1])
        simulator = FarFieldSimulator(
            is_torch=True,
            **simulation_config,
        )

        super().__init__(
            measured_dir=data_dir,
            original_dir=os.path.join(celeba_root, "celeba", "img_align_celeba"),
            simulator=simulator,
            measurement_ext="png",
            original_ext="jpg",
            downsample=downsample,
            background=background,
            flip=False,  # will do flipping only on measurement
            **kwargs,
        )

    def _get_images_pair(self, idx):

        # more standard image formats: png, jpg, tiff, etc.
        lensless_fp = os.path.join(self.measured_dir, self.files[idx])
        original_fp = os.path.join(self.original_dir, self.files[idx][:-3] + self.original_ext)
        lensless = load_image(
            lensless_fp, downsample=self.pre_downsample, flip=self.flip_measurement
        )
        original = load_image(original_fp[:-3] + self.original_ext)

        # convert to float
        if lensless.dtype == np.uint8:
            lensless = lensless.astype(np.float32) / 255
            original = original.astype(np.float32) / 255
        else:
            # 16 bit
            lensless = lensless.astype(np.float32) / 65535
            original = original.astype(np.float32) / 65535

        # convert to torch
        lensless = torch.from_numpy(lensless)
        original = torch.from_numpy(original)

        # project original image to lensed space
        with torch.no_grad():
            lensed = self.sim.propagate_image(original, return_object_plane=True)

        if self.vertical_shift is not None:
            lensed = torch.roll(lensed, self.vertical_shift, dims=-3)
        if self.horizontal_shift is not None:
            lensed = torch.roll(lensed, self.horizontal_shift, dims=-2)

        return lensless, lensed


class MeasuredDataset(DualDataset):
    """
    Dataset consisting of lensless and corresponding lensed image.
    It can be used with a PyTorch DataLoader to load a batch of lensless and corresponding lensed images.
    Unless the setup is perfectly calibrated, one should expect to have to use ``transform_lensed`` to adjust the alignment and rotation.
    """

    def __init__(
        self,
        root_dir,
        lensless_fn="diffuser",
        lensed_fn="lensed",
        image_ext="npy",
        **kwargs,
    ):
        """
        Dataset consisting of lensless and corresponding lensed image. Default parameters are for the
        `DiffuserCam Lensless Mirflickr Dataset (DLMD) <https://waller-lab.github.io/LenslessLearning/dataset.html>`_.

        Parameters
        ----------
        root_dir : str
            Path to the test dataset. It is expected to contain two folders: ones of lensless images and one of lensed images.
        lensless_fn : str, optional
            Name of the folder containing the lensless images, by default "diffuser".
        lensed_fn : str, optional
            Name of the folder containing the lensed images, by default "lensed".
        image_ext : str, optional
            Extension of the images, by default "npy".
        """

        super(MeasuredDataset, self).__init__(**kwargs)

        self.root_dir = root_dir
        self.lensless_dir = os.path.join(root_dir, lensless_fn)
        self.lensed_dir = os.path.join(root_dir, lensed_fn)
        assert os.path.isdir(self.lensless_dir)
        assert os.path.isdir(self.lensed_dir)

        self.image_ext = image_ext.lower()

        files = natural_sort(glob.glob(os.path.join(self.lensless_dir, "*." + image_ext)))
        self.files = [os.path.basename(fn) for fn in files]

        if len(self.files) == 0:
            raise FileNotFoundError(
                f"No files found in {self.lensless_dir} with extension {image_ext}"
            )

    def __len__(self):
        if self.indices is None:
            return len(self.files)
        else:
            return len([i for i in self.indices if i < len(self.files)])

    def _get_images_pair(self, idx):
        if self.image_ext == "npy" or self.image_ext == "npz":
            lensless_fp = os.path.join(self.lensless_dir, self.files[idx])
            lensed_fp = os.path.join(self.lensed_dir, self.files[idx])
            lensless = np.load(lensless_fp)
            lensed = np.load(lensed_fp)

        else:
            # more standard image formats: png, jpg, tiff, etc.
            lensless_fp = os.path.join(self.lensless_dir, self.files[idx])
            lensed_fp = os.path.join(self.lensed_dir, self.files[idx])
            lensless = load_image(lensless_fp)
            lensed = load_image(lensed_fp)

            # convert to float
            if lensless.dtype == np.uint8:
                lensless = lensless.astype(np.float32) / 255
                lensed = lensed.astype(np.float32) / 255
            else:
                # 16 bit
                lensless = lensless.astype(np.float32) / 65535
                lensed = lensed.astype(np.float32) / 65535

        return lensless, lensed


class DiffuserCamMirflickr(MeasuredDataset):
    """
    Helper class for DiffuserCam Mirflickr dataset.

    Note that image colors are in BGR format: https://github.com/Waller-Lab/LenslessLearning/blob/master/utils.py#L432
    """

    def __init__(
        self,
        dataset_dir,
        psf_path,
        downsample=2,
        **kwargs,
    ):

        # check psf path exist
        if not os.path.exists(psf_path):
            psf_path = os.path.join(
                os.path.dirname(__file__), "..", "..", "data", "psf", "diffusercam_psf.tiff"
            )

            try:
                from torchvision.datasets.utils import download_url
            except ImportError:
                exit()
            msg = "Do you want to download the DiffuserCam PSF (5.9MB)?"

            # default to yes if no input is given
            valid = input("%s (Y/n) " % msg).lower() != "n"
            output_path = os.path.join(os.path.dirname(__file__), "..", "..", "data", "psf")
            if valid:
                url = "https://drive.switch.ch/index.php/s/BteiuEcONmhmDSn/download"
                filename = "diffusercam_psf.tiff"
                download_url(url, output_path, filename=filename)

        psf, background = load_psf(
            psf_path,
            downsample=downsample * 4,  # PSF is 4x the resolution of the images
            return_float=True,
            return_bg=True,
            bg_pix=(0, 15),
        )
        transform_BRG2RGB = transforms.Lambda(lambda x: x[..., [2, 1, 0]])
        self.psf = transform_BRG2RGB(torch.from_numpy(psf))
        self.allowed_idx = np.arange(2, 25001)

        assert os.path.isdir(os.path.join(dataset_dir, "diffuser_images")) and os.path.isdir(
            os.path.join(dataset_dir, "ground_truth_lensed")
        ), "Dataset should contain 'diffuser_images' and 'ground_truth_lensed' folders. It can be downloaded from https://waller-lab.github.io/LenslessLearning/dataset.html"

        super().__init__(
            root_dir=dataset_dir,
            background=background,
            downsample=downsample,
            flip=False,
            transform_lensless=transform_BRG2RGB,
            transform_lensed=transform_BRG2RGB,
            lensless_fn="diffuser_images",
            lensed_fn="ground_truth_lensed",
            image_ext="npy",
            **kwargs,
        )

    def _get_images_pair(self, idx):

        assert idx >= self.allowed_idx.min(), f"idx should be >= {self.allowed_idx.min()}"
        assert idx <= self.allowed_idx.max(), f"idx should be <= {self.allowed_idx.max()}"

        fn = f"im{idx}.npy"
        lensless_fp = os.path.join(self.lensless_dir, fn)
        lensed_fp = os.path.join(self.lensed_dir, fn)
        lensless = np.load(lensless_fp)
        lensed = np.load(lensed_fp)

        return lensless, lensed


class DiffuserCamTestDataset(MeasuredDataset):
    """
    Dataset consisting of lensless and corresponding lensed image. This is the standard dataset used for benchmarking.
    """

    def __init__(
        self,
        data_dir=None,
        n_files=None,
        downsample=2,
    ):
        """
        Dataset consisting of lensless and corresponding lensed image. Default parameters are for the test set of
        `DiffuserCam Lensless Mirflickr Dataset (DLMD) <https://waller-lab.github.io/LenslessLearning/dataset.html>`_.

        Parameters
        ----------
        data_dir : str, optional
            The path to ``DiffuserCam_Test`` dataset, by default looks inside the ``data`` folder.
        n_files : int, optional
            Number of image pairs to load in the dataset , by default use all.
        downsample : int, optional
            Downsample factor of the lensless images, by default 2. Note that the PSF has a resolution of 4x of the images.
        """

        # download dataset if necessary
        if data_dir is None:
            data_dir = os.path.join(
                os.path.dirname(__file__), "..", "..", "data", "DiffuserCam_Test"
            )
        if not os.path.isdir(data_dir):
            main_dir = os.path.join(os.path.dirname(__file__), "..", "..", "data")
            print("DiffuserCam test set not found for benchmarking.")
            try:
                from torchvision.datasets.utils import download_and_extract_archive
            except ImportError:
                exit()
            msg = "Do you want to download the dataset (3.5GB)?"

            # default to yes if no input is given
            valid = input("%s (Y/n) " % msg).lower() != "n"
            if valid:
                url = "https://drive.switch.ch/index.php/s/D3eRJ6PRljfHoH8/download"
                filename = "DiffuserCam_Test.zip"
                download_and_extract_archive(url, main_dir, filename=filename, remove_finished=True)

        psf_fp = os.path.join(data_dir, "psf.tiff")
        psf, background = load_psf(
            psf_fp,
            downsample=downsample * 4,  # PSF is 4x the resolution of the images
            return_float=True,
            return_bg=True,
            bg_pix=(0, 15),
            flip_ud=True,
            flip_lr=False,
        )

        # transform from BGR to RGB
        transform_BRG2RGB = transforms.Lambda(lambda x: x[..., [2, 1, 0]])

        self.psf = transform_BRG2RGB(torch.from_numpy(psf))

        if n_files is None:
            indices = None
        else:
            indices = range(n_files)

        super().__init__(
            root_dir=data_dir,
            indices=indices,
            background=background,
            downsample=downsample,
            flip=False,
            flip_ud=True,
            flip_lr=False,
            transform_lensless=transform_BRG2RGB,
            transform_lensed=transform_BRG2RGB,
            lensless_fn="diffuser",
            lensed_fn="lensed",
            image_ext="npy",
        )


class SimulatedDatasetTrainableMask(SimulatedFarFieldDataset):
    """
    Dataset of propagated images (through simulation) from a Torch Dataset with learnable mask.
    The `waveprop <https://github.com/ebezzam/waveprop/blob/master/waveprop/simulation.py>`_ package is used for the simulation,
    assuming a far-field propagation and a shift-invariant system with a single point spread function (PSF).
    To ensure autograd compatibility, the dataloader should have ``num_workers=0``.
    """

    def __init__(
        self,
        mask,
        dataset,
        simulator,
        **kwargs,
    ):
        """
        Parameters
        ----------

        mask : :py:class:`lensless.hardware.trainable_mask.TrainableMask`
            Mask to use for simulation. Should be a 4D tensor with shape [1, H, W, C]. Simulation of multi-depth data is not supported yet.
        dataset : :py:class:`torch.utils.data.Dataset`
            Dataset to propagate. Should output images with shape [H, W, C] unless ``dataset_is_CHW`` is ``True`` (and therefore images have the dimension ordering of [C, H, W]).
        simulator : :py:class:`lensless.utils.simulation.FarFieldSimulator`
            Simulator object used on images from ``dataset``. Waveprop simulator to use for the simulation. It is expected to have ``is_torch = True``.
        """

        self._mask = mask

        temp_psf = self._mask.get_psf()
        test_sim = FarFieldSimulator(psf=temp_psf, **simulator.params)
        assert (
            test_sim.conv_dim == simulator.conv_dim
        ).all(), "PSF shape should match simulator shape"
        assert (
            not simulator.quantize
        ), "Simulator should not perform quantization to maintain differentiability. Please set quantize=False"

        super(SimulatedDatasetTrainableMask, self).__init__(dataset, simulator, **kwargs)

    def set_psf(self, psf=None):
        """
        Set the PSF of the simulator.

        Parameters
        ----------
        psf : :py:class:`torch.Tensor`, optional
            PSF to use for the simulation. If ``None``, the PSF of the mask is used.
        """
        if psf is None:
            psf = self._mask.get_psf()
        self.sim.set_point_spread_function(psf)


class HITLDatasetTrainableMask(SimulatedDatasetTrainableMask):
    """
    Dataset of on-the-fly measurements and simulated ground-truth.
    """

    def __init__(
        self,
        rpi_username,
        rpi_hostname,
        celeba_root,
        display_config,
        capture_config,
        mask_center,
        **kwargs,
    ):
        self.rpi_username = rpi_username
        self.rpi_hostname = rpi_hostname
        self.celeba_root = celeba_root
        assert os.path.isdir(self.celeba_root)

        self.display_config = display_config
        self.capture_config = capture_config
        self.mask_center = mask_center

        super(HITLDatasetTrainableMask, self).__init__(**kwargs)

    def __getitem__(self, index):

        # propagate through mask in digital model
        _, lensed = super().__getitem__(index)

        ## measure lensless image
        # get image file path
        idx = self.dataset.indices[index]

        # twice nested as we do train-test split of subset of CelebA
        fn = self.dataset.dataset.dataset.filename[idx]
        fp = os.path.join(self.celeba_root, "celeba", "img_align_celeba", fn)

        # display on screen
        display(
            fp=fp,
            rpi_username=self.rpi_username,
            rpi_hostname=self.rpi_hostname,
            **self.display_config,
        )

        # set mask
        with torch.no_grad():
            subpattern = self._mask.get_vals()
            subpattern_np = subpattern.detach().cpu().numpy().copy()
            pattern = adafruit_sub2full(
                subpattern_np,
                center=self.mask_center,
            )
        set_programmable_mask(
            pattern,
            self._mask.device,
            self.rpi_username,
            self.rpi_hostname,
        )

        # take picture
        _, img = capture(
            rpi_username=self.rpi_username,
            rpi_hostname=self.rpi_hostname,
            verbose=False,
            **self.capture_config,
        )

        # -- normalize
        img = img.astype(np.float32) / img.max()

        # prep
        img = torch.from_numpy(img)
        # -- if [H, W, C] -> [D, H, W, C]
        if len(img.shape) == 3:
            img = img.unsqueeze(0)

        if self.background is not None:
            img = img - self.background

        # flip image x and y if needed
        if self.capture_config.flip:
            img = torch.rot90(img, dims=(-3, -2), k=2)

        # return simulated images (replace simulated with measured)
        return img, lensed


class DiffuserCamMirflickrHF(DualDataset):
    def __init__(
        self,
        split,
        repo_id="bezzam/DiffuserCam-Lensless-Mirflickr-Dataset",
        psf="psf.tiff",
        downsample=2,
        flip_ud=True,
        dtype="float32",
        **kwargs,
    ):
        """
        Parameters
        ----------
        split : str
            Split of the dataset to use: 'train', 'test', or 'all'.
        downsample : int, optional
            Downsample factor of the PSF, which is 4x the resolution of the images, by default 6 for resolution of (180, 320).
        flip_ud : bool, optional
            If True, data is flipped up-down, by default ``True``. Otherwise data is upside-down.
        """

        # get dataset
        self.dataset = load_dataset(repo_id, split=split)

        # get PSF
        psf_fp = hf_hub_download(repo_id=repo_id, filename=psf, repo_type="dataset")
        psf, bg = load_psf(
            psf_fp,
            verbose=False,
            downsample=downsample * 4,
            return_bg=True,
            flip_ud=flip_ud,
            dtype=dtype,
            bg_pix=(0, 15),
        )
        self.psf = torch.from_numpy(psf)

        super(DiffuserCamMirflickrHF, self).__init__(
            flip_ud=flip_ud, downsample=downsample, background=bg, **kwargs
        )

    def __len__(self):
        return len(self.dataset)

    def _get_images_pair(self, idx):
        lensless = np.array(self.dataset[idx]["lensless"])
        lensed = np.array(self.dataset[idx]["lensed"])

        # normalize
        lensless = lensless.astype(np.float32) / 255
        lensed = lensed.astype(np.float32) / 255

        return lensless, lensed


class HFSimulated(DualDataset):
    def __init__(
        self,
        huggingface_repo,
        split,
        n_files=None,
        psf=None,
        downsample=1,
        cache_dir=None,
        single_channel_psf=False,
        flipud=False,
        display_res=None,
        alignment=None,
        sensor="rpi_hq",
        slm="adafruit",
        simulation_config=dict(),
        snr_db=40,
        **kwargs,
    ):
        """
        Wrapper for Hugging Face datasets, where lensless images are simulated from lensed ones.

        This is used for seeing how simulated lensless images compare with real ones.
        """

        if isinstance(split, str):
            if n_files is not None:
                split = f"{split}[0:{n_files}]"
            self.dataset = load_dataset(huggingface_repo, split=split, cache_dir=cache_dir)
        elif isinstance(split, Dataset):
            self.dataset = split
        else:
            raise ValueError("split should be a string or a Dataset object")

        # deduce downsampling factor from the first image
        data_0 = self.dataset[0]
        self.downsample = downsample
        # -- use lensless data just for shape but using lensed data in simulation
        lensless = np.array(data_0["lensless"])
        self.lensless_shape = np.array(lensless.shape[:2]) // self.downsample

        # download PSF from huggingface
        # TODO : assuming psf is not None
        self.multimask = False
        self.convolver = None
        if psf is not None:
            psf_fp = hf_hub_download(repo_id=huggingface_repo, filename=psf, repo_type="dataset")
            psf, _ = load_psf(
                psf_fp,
                shape=self.lensless_shape,
                return_float=True,
                return_bg=True,
                flip_ud=flipud,
                bg_pix=(0, 15),
                single_psf=single_channel_psf,
            )
            self.psf = torch.from_numpy(psf)
            if single_channel_psf:
                # replicate across three channels
                self.psf = self.psf.repeat(1, 1, 1, 3)

            # create convolver object
            self.convolver = RealFFTConvolve2D(self.psf)

        elif "mask_label" in data_0:
            self.multimask = True
            mask_labels = []
            for i in range(len(self.dataset)):
                mask_labels.append(self.dataset[i]["mask_label"])
            mask_labels = list(set(mask_labels))

            # simulate all PSFs
            self.psf = dict()
            for label in mask_labels:
                mask_fp = hf_hub_download(
                    repo_id=huggingface_repo,
                    filename=f"masks/mask_{label}.npy",
                    repo_type="dataset",
                )
                mask_vals = np.load(mask_fp)

                if psf is None:
                    sensor_res = sensor_dict[sensor][SensorParam.RESOLUTION]
                    downsample_fact = min(sensor_res / lensless.shape[:2])
                else:
                    downsample_fact = 1

                mask = AdafruitLCD(
                    initial_vals=torch.from_numpy(mask_vals.astype(np.float32)),
                    sensor=sensor,
                    slm=slm,
                    downsample=downsample_fact,
                    flipud=rotate or flipud,  # TODO separate commands?
                    use_waveprop=simulation_config.get("use_waveprop", False),
                    scene2mask=simulation_config.get("scene2mask", None),
                    mask2sensor=simulation_config.get("mask2sensor", None),
                    deadspace=simulation_config.get("deadspace", True),
                )
                self.psf[label] = mask.get_psf().detach()

                assert (
                    self.psf[label].shape[-3:-1] == lensless.shape[:2]
                ), f"PSF shape should match lensless shape: PSF {self.psf[label].shape[-3:-1]} vs lensless {lensless.shape[:2]}"

            # create convolver object
            self.convolver = RealFFTConvolve2D(self.psf[label])
        assert self.convolver is not None

        self.crop = None
        self.random_flip = None
        self.flipud = flipud
        self.snr_db = snr_db

        self.display_res = display_res
        self.alignment = None
        self.cropped_lensed_shape = None
        if alignment is not None:
            self.alignment = dict(alignment.copy())
            self.alignment["top_left"] = (
                int(self.alignment["top_left"][0] / downsample),
                int(self.alignment["top_left"][1] / downsample),
            )
            self.alignment["height"] = int(self.alignment["height"] / downsample)

            original_aspect_ratio = display_res[1] / display_res[0]
            self.alignment["width"] = int(self.alignment["height"] * original_aspect_ratio)
            self.cropped_lensed_shape = (self.alignment["height"], self.alignment["width"], 3)

        super(HFSimulated, self).__init__(**kwargs)

    def __len__(self):
        return len(self.dataset)

    def _get_images_pair(self, idx):

        # load image
        lensed_np = np.array(self.dataset[idx]["lensed"])
        if self.flipud:
            lensed_np = np.flipud(lensed_np)

        # convert to float
        if lensed_np.dtype == np.uint8:
            lensed_np = lensed_np.astype(np.float32) / 255
        else:
            # 16 bit
            lensed_np = lensed_np.astype(np.float32) / 65535

        # resize if necessary
        if self.cropped_lensed_shape is not None:
            cropped_lensed_np = resize(
                lensed_np, shape=self.cropped_lensed_shape, interpolation=cv2.INTER_NEAREST
            )
            lensed_np = np.zeros(tuple(self.lensless_shape) + (3,), dtype=np.float32)
            lensed_np[
                self.alignment["top_left"][0] : self.alignment["top_left"][0]
                + self.alignment["height"],
                self.alignment["top_left"][1] : self.alignment["top_left"][1]
                + self.alignment["width"],
            ] = cropped_lensed_np

        elif (self.lensless_shape != np.array(lensed_np.shape[:2])).any():

            lensed_np = resize(
                lensed_np, shape=self.lensless_shape, interpolation=cv2.INTER_NEAREST
            )
        lensed = torch.from_numpy(lensed_np)

        # simulate lensless with convolution
        lensed = lensed.unsqueeze(0)  # add batch dimension

        if self.multimask:
            mask_label = self.dataset[idx]["mask_label"]
            self.convolver.set_psf(self.psf[mask_label])
        lensless = self.convolver.convolve(lensed)

        # add noise
        if self.snr_db is not None:
            lensless = add_shot_noise(lensless, self.snr_db)

        if lensless.max() > 1:
            print("CLIPPING!")
            lensless /= lensless.max()

        if self.cropped_lensed_shape:
            return lensless, torch.from_numpy(cropped_lensed_np)
        else:
            return lensless, lensed

    def __getitem__(self, idx):
        lensless, lensed = super().__getitem__(idx)
        if self.multimask:
            mask_label = self.dataset[idx]["mask_label"]
            return lensless, lensed, self.psf[mask_label]
        return lensless, lensed

    def extract_roi(self, reconstruction, lensed=None, axis=(1, 2), **kwargs):
        """
        Extract region of interest from lensless and lensed images.
        """

        n_dim = len(reconstruction.shape)
        assert max(axis) < n_dim, "Axis should be within the dimensions of the reconstruction."

        # add batch dimension
        if n_dim == 3:
            if isinstance(reconstruction, torch.Tensor):
                reconstruction = reconstruction.unsqueeze(0)
            else:
                reconstruction = reconstruction[np.newaxis]
            # increment axis
            axis = (axis[0] + 1, axis[1] + 1)

        # extract
        if self.alignment is not None:
            top_left = self.alignment["top_left"]
            height = self.alignment["height"]
            width = self.alignment["width"]

            # extract according to axis
            index = [slice(None)] * n_dim
            index[axis[0]] = slice(top_left[0], top_left[0] + height)
            index[axis[1]] = slice(top_left[1], top_left[1] + width)
            reconstruction = reconstruction[tuple(index)]

            # rotate if necessary
            angle = self.alignment.get("angle", 0)
            if isinstance(reconstruction, torch.Tensor) and angle:
                reconstruction = F.rotate(reconstruction, angle, expand=False)
            elif angle:
                reconstruction = rotate(reconstruction, angle, axes=axis, reshape=False)

        # remove batch dimension
        if n_dim == 3:
            if isinstance(reconstruction, torch.Tensor):
                reconstruction = reconstruction.squeeze(0)
            else:
                reconstruction = reconstruction[0]

        if lensed is None:
            return reconstruction
        return reconstruction, lensed


class HFDataset(Dataset):
    def __init__(
        self,
        huggingface_repo,
        split,
        n_files=None,
        psf=None,
        rotate=False,  # just the lensless image
        flipud=False,
        flip_lensed=False,
        downsample=1,
        downsample_lensed=1,
        input_snr=None,
        psf_snr=None,
        display_res=None,
        sensor="rpi_hq",
        slm="adafruit",
        alignment=None,
        return_mask_label=False,
        save_psf=False,
        simulation_config=dict(),
        simulate_lensless=False,
        force_rgb=False,
        cache_dir=None,
        single_channel_psf=False,
        random_flip=False,
        bg_snr_range=None,
        bg_fp=None,
        **kwargs,
    ):
        """
        Wrapper for lensless datasets on Hugging Face.

        Parameters
        ----------
        huggingface_repo : str
            Hugging Face repository ID.
        split : str or :py:class:`torch.utils.data.Dataset`
            Split of the dataset to use: 'train', 'test', or 'all'. If a Dataset object is given, it is used directly.
        n_files : int, optional
            Number of files to load from the dataset, by default None, namely all.
        psf : str, optional
            File name of the PSF at the repository. If None, it is assumed that there is a mask pattern from which the PSF is simulated, by default None.
        rotate : bool, optional
            If True, lensless images and PSF are rotated 180 degrees. Lensed/original image is not rotated! By default False.
        downsample : float, optional
            Downsample factor of the lensless images, by default 1.
        downsample_lensed : float, optional
            Downsample factor of the lensed images, by default 1.
        display_res : tuple, optional
            Resolution of images when displayed on screen during measurement.
        sensor : str, optional
            If `psf` not provided, the sensor to use for the PSF simulation, by default "rpi_hq".
        slm : str, optional
            If `psf` not provided, the SLM to use for the PSF simulation, by default "adafruit".
        alignment : dict, optional
            Alignment parameters between lensless and lensed data.
            If "top_left", "height", and "width" are provided, the region-of-interest from the reconstruction of ``lensless`` is extracted and ``lensed`` is reshaped to match.
            If "crop" is provided, the region-of-interest is extracted from the simulated lensed image, namely a ``simulation`` configuration should be provided within ``alignment``.
        return_mask_label : bool, optional
            If multimask dataset, return the mask label (True) or the corresponding PSF (False).
        save_psf : bool, optional
            If multimask dataset, save the simulated PSFs.
        random_flip : bool, optional
            If True, randomly flip the lensless images vertically and horizonally with equal probability. By default, no flipping.
        simulation_config : dict, optional
            Simulation parameters for PSF if using a mask pattern.
        bg_snr_range : list, optional
            List [low, high] of range of possible SNRs for which to add the background. Used in conjunction with 'bg'.
        bg_fp : string, optional
            File path of background to add to the data for simulating a measurement in ambient light.

        """

        if isinstance(split, str):
            if n_files is not None:
                split = f"{split}[0:{n_files}]"
            self.dataset = load_dataset(huggingface_repo, split=split, cache_dir=cache_dir)
        elif isinstance(split, Dataset):
            self.dataset = split
        else:
            raise ValueError("split should be a string or a Dataset object")

        self.rotate = rotate
        self.flipud = flipud
        self.flip_lensed = flip_lensed
        self.display_res = display_res
        self.return_mask_label = return_mask_label
        self.force_rgb = force_rgb  # if some data is not 3D
        self.sensor = sensor
        self.slm = slm
        self.simulation_config = simulation_config
        self.input_snr = input_snr

        # augmentation
        self.random_flip = random_flip

        # deduce downsampling factor from the first image
        data_0 = self.dataset[0]
        self.downsample_lensless = downsample
        self.downsample_lensed = downsample_lensed
        lensless = np.array(data_0["lensless"])
        if "ambient" in data_0.keys():
            self.measured_bg = True
        else:
            self.measured_bg = False

        if self.downsample_lensless != 1.0:
            lensless = resize(lensless, factor=1 / self.downsample_lensless)
        if psf is None:
            sensor_res = sensor_dict[sensor][SensorParam.RESOLUTION]
            downsample_fact = min(sensor_res / lensless.shape[:2])
        else:
            downsample_fact = 1
        self.downsample_fact = downsample_fact

        # deduce recon shape from original image
        self.alignment = None
        self.crop = None
        if alignment is not None:
            if alignment.get("top_left", None) is not None:
                self.alignment = dict(alignment.copy())
                self.alignment["top_left"] = (
                    int(self.alignment["top_left"][0] / downsample),
                    int(self.alignment["top_left"][1] / downsample),
                )
                self.alignment["height"] = int(self.alignment["height"] / downsample)

                original_aspect_ratio = display_res[1] / display_res[0]
                self.alignment["width"] = int(self.alignment["height"] * original_aspect_ratio)

            if alignment.get("topright", None) is not None:
                # typo in original configuration
                self.alignment = dict(alignment.copy())
                self.alignment["top_left"] = (
                    int(self.alignment["topright"][0] / downsample),
                    int(self.alignment["topright"][1] / downsample),
                )
                self.alignment["height"] = int(self.alignment["height"] / downsample)

                original_aspect_ratio = display_res[1] / display_res[0]
                self.alignment["width"] = int(self.alignment["height"] * original_aspect_ratio)

            elif alignment.get("crop", None) is not None:
                self.crop = dict(alignment["crop"].copy())
                self.crop["vertical"][0] = int(self.crop["vertical"][0] / downsample)
                self.crop["vertical"][1] = int(self.crop["vertical"][1] / downsample)
                self.crop["horizontal"][0] = int(self.crop["horizontal"][0] / downsample)
                self.crop["horizontal"][1] = int(self.crop["horizontal"][1] / downsample)

        # download all masks
        # TODO: reshape directly with lensless image shape
        self.multimask = False
        self.huggingface_repo = huggingface_repo
        if psf is not None:
            # download PSF from huggingface
            psf_fp = hf_hub_download(repo_id=huggingface_repo, filename=psf, repo_type="dataset")
            psf, _ = load_psf(
                psf_fp,
                shape=lensless.shape,
                return_float=True,
                return_bg=True,
                flip=self.rotate,
                flip_ud=flipud,
                bg_pix=(0, 15),
                force_rgb=force_rgb,
                single_psf=single_channel_psf,
            )
            self.psf = torch.from_numpy(psf)
            if single_channel_psf:
                # replicate across three channels
                self.psf = self.psf.repeat(1, 1, 1, 3)

            if psf_snr is not None:
                # # add noise to PSF
                # self.psf = add_shot_noise(self.psf, psf_snr)
                # add Gaussian noise to PSF
                noise = torch.randn_like(self.psf)
                noise_var = torch.var(noise.flatten())
                psf_var = torch.var(self.psf.flatten())
                noise *= torch.sqrt(psf_var / noise_var) / 10 ** (psf_snr / 20)
                self.psf += noise

                # save PSF as torch tensor for loading model later on
                torch.save(self.psf, "psf.pt")

            if save_psf:
                # same viewable image of PSF
                save_image(self.psf.squeeze().cpu().numpy(), f"{split}_psf.png")

        elif "mask_label" in data_0:
            self.multimask = True
            mask_labels = []
            for i in range(len(self.dataset)):
                mask_labels.append(self.dataset[i]["mask_label"])
            mask_labels = list(set(mask_labels))
            self.mask_labels = mask_labels

            # simulate all PSFs
            self.psf = dict()
            for label in mask_labels:

                mask_vals = self.get_mask_vals(label)
                self.psf[label] = self.simulate_psf(mask_vals)

                assert (
                    self.psf[label].shape[-3:-1] == lensless.shape[:2]
                ), f"PSF shape should match lensless shape: PSF {self.psf[label].shape[-3:-1]} vs lensless {lensless.shape[:2]}"

                if save_psf:
                    # same viewable image of PSF
                    save_image(self.psf[label].squeeze().cpu().numpy(), f"psf_{label}.png")

        else:

            # single mask pattern
            mask_fp = hf_hub_download(
                repo_id=huggingface_repo, filename="mask_pattern.npy", repo_type="dataset"
            )
            mask_vals = np.load(mask_fp)
            self.psf = self.simulate_psf(mask_vals)
            assert (
                self.psf.shape[-3:-1] == lensless.shape[:2]
            ), "PSF shape should match lensless shape"

            if save_psf:
                # same viewable image of PSF
                save_image(self.psf.squeeze().cpu().numpy(), "psf.png")

        # create simulator
        self.simulate_lensless = simulate_lensless
        if simulate_lensless:
            assert (
                alignment is not None and alignment.get("simulation") is not None
            ), "Need simulation parameters for lensless images"
        self.simulator = None
        if alignment is not None and "simulation" in alignment:
            simulation_config = dict(alignment["simulation"].copy())
            simulation_config["output_dim"] = tuple(self.psf.shape[-3:-1])
            if simulation_config.get("vertical_shift", None) is not None:
                simulation_config["vertical_shift"] = int(
                    simulation_config["vertical_shift"] / downsample
                )
            if simulation_config.get("horizontal_shift", None) is not None:
                simulation_config["horizontal_shift"] = int(
                    simulation_config["horizontal_shift"] / downsample
                )
            simulator = FarFieldSimulator(
                psf=self.psf if self.simulate_lensless else None,
                is_torch=True,
                **simulation_config,
            )
            self.simulator = simulator

        if bg_fp is not None:
            assert (
                bg_snr_range is not None
            ), "Since a background path was provided, the SNR range should not be empty"
            bg = load_image(
                bg_fp,
                shape=lensless.shape,
                return_float=True,
                flip=rotate,
            )
            self.bg_sim = torch.from_numpy(bg)
            # Used for background noise addition
            self.bg_snr_range = bg_snr_range
            # Precomputing for efficiency (used in the SNR computations)
            self.background_var = torch.var(self.bg_sim.flatten())
        else:
            self.bg_sim = None
            self.bg_snr_range = None
            self.background_var = None

    def __len__(self):
        return len(self.dataset)

    def get_mask_vals(self, idx):
        assert self.multimask
        assert idx in self.mask_labels
        mask_fp = hf_hub_download(
            repo_id=self.huggingface_repo,
            filename=f"masks/mask_{idx}.npy",
            repo_type="dataset",
        )
        return np.load(mask_fp)

    def simulate_psf(self, mask_vals):
        mask = AdafruitLCD(
            initial_vals=torch.from_numpy(mask_vals.astype(np.float32)),
            sensor=self.sensor,
            slm=self.slm,
            downsample=self.downsample_fact,
            flipud=self.rotate or self.flipud,  # TODO separate commands?
            use_waveprop=self.simulation_config.get("use_waveprop", False),
            scene2mask=self.simulation_config.get("scene2mask", None),
            mask2sensor=self.simulation_config.get("mask2sensor", None),
            deadspace=self.simulation_config.get("deadspace", True),
        )
        return mask.get_psf().detach()

    def _get_images_pair(self, idx):

        # load images
        lensless_np = np.array(self.dataset[idx]["lensless"])
        lensed_np = np.array(self.dataset[idx]["lensed"])
        background_np = np.array(self.dataset[idx]["ambient"]) if self.measured_bg else None

        if self.force_rgb:
            if len(lensless_np.shape) == 2:
                warnings.warn(f"Converting lensless[{idx}] to RGB")
                lensless_np = np.stack([lensless_np] * 3, axis=2)
            elif len(lensless_np.shape) == 3:
                pass
            else:
                raise ValueError(f"lensless[{idx}] should be 2D or 3D")

            if len(lensed_np.shape) == 2:
                warnings.warn(f"Converting lensed[{idx}] to RGB")
                lensed_np = np.stack([lensed_np] * 3, axis=2)
            elif len(lensed_np.shape) == 3:
                pass

            if len(background_np.shape) == 2:
                warnings.warn(f"Converting background[{idx}] to RGB")
                background_np = (
                    np.stack([background_np] * 3, axis=2) if background_np is not None else None
                )
            elif len(background_np.shape) == 3:
                pass

        # convert to float
        if lensless_np.dtype == np.uint8:
            lensless_np = lensless_np.astype(np.float32) / 255
            lensed_np = lensed_np.astype(np.float32) / 255
            background_np = (
                background_np.astype(np.float32) / 255 if background_np is not None else None
            )
        else:
            # 16 bit
            lensless_np = lensless_np.astype(np.float32) / 65535
            lensed_np = lensed_np.astype(np.float32) / 65535
            background_np = (
                background_np.astype(np.float32) / 65535 if background_np is not None else None
            )

        # downsample if necessary
        if self.downsample_lensless != 1.0:
            lensless_np = resize(
                lensless_np, factor=1 / self.downsample_lensless, interpolation=cv2.INTER_NEAREST
            )
            background_np = (
                resize(
                    background_np,
                    factor=1 / self.downsample_lensless,
                    interpolation=cv2.INTER_NEAREST,
                )
                if background_np is not None
                else None
            )

        lensless = lensless_np
        lensed = lensed_np
        background = background_np if background_np is not None else None

        if self.simulator is not None:
            # convert to torch
            lensless = torch.from_numpy(lensless_np)
            lensed = torch.from_numpy(lensed_np)

            # project original image to lensed space
            with torch.no_grad():

                if self.simulate_lensless:
                    lensless, lensed = self.simulator.propagate_image(
                        lensed, return_object_plane=True
                    )
                else:
                    lensed = self.simulator.propagate_image(lensed, return_object_plane=True)

        elif self.alignment is not None:
            lensed = resize(
                lensed_np,
                shape=(self.alignment["height"], self.alignment["width"], 3),
                interpolation=cv2.INTER_NEAREST,
            )
        elif self.display_res is not None:
            lensed = resize(
                lensed_np, shape=(*self.display_res, 3), interpolation=cv2.INTER_NEAREST
            )
        elif self.downsample_lensed != 1.0:
            lensed = resize(
                lensed_np,
                factor=1 / self.downsample_lensed,
                interpolation=cv2.INTER_NEAREST,
            )

        return lensless, lensed, background if background is not None else None

    def __getitem__(self, idx):

        lensless, lensed, background = self._get_images_pair(idx)

        if isinstance(lensless, np.ndarray):
            # to torch
            lensless = torch.from_numpy(lensless)
            lensed = torch.from_numpy(lensed)
            background = torch.from_numpy(background) if background is not None else None
        # If [H, W, C] -> [D, H, W, C]
        if len(lensless.shape) == 3:
            lensless = lensless.unsqueeze(0)
        if len(lensed.shape) == 3:
            lensed = lensed.unsqueeze(0)
        if background is not None and len(background.shape) == 3:
            background = background.unsqueeze(0)

        if self.input_snr is not None:
            lensless = add_shot_noise(lensless, self.input_snr)

        if not self.simulate_lensless:  # TODO apply transformation to bg as well?
            if self.rotate:
                lensless = torch.rot90(lensless, dims=(-3, -2), k=2)
            if self.flipud:
                lensless = torch.flip(lensless, dims=(-3,))

        if self.flip_lensed:
            if self.rotate:
                lensed = torch.rot90(lensed, dims=(-3, -2), k=2)
            if self.flipud:
                lensed = torch.flip(lensed, dims=(-3,))

        if self.multimask:
            mask_label = self.dataset[idx]["mask_label"]

        flip_lr = False
        flip_ud = False
        if self.random_flip:
            flip_lr = torch.rand(1) > 0.5
            flip_ud = torch.rand(1) > 0.5

            if self.multimask:
                psf_aug = self.psf[mask_label].clone()
            else:
                psf_aug = self.psf.clone()

            if flip_lr:
                lensless = torch.flip(lensless, dims=(-2,))
                lensed = torch.flip(lensed, dims=(-2,))
                psf_aug = torch.flip(psf_aug, dims=(-2,))
                background = torch.flip(background, dims=(-2,))
            if flip_ud:
                lensless = torch.flip(lensless, dims=(-3,))
                lensed = torch.flip(lensed, dims=(-3,))
                psf_aug = torch.flip(psf_aug, dims=(-3,))
                background = torch.flip(background, dims=(-3,))

        return_items = [lensless, lensed]
        if self.multimask:
            if self.return_mask_label:
                return_items.append(mask_label)
            else:
                return_items.append(self.psf[mask_label])
            if self.random_flip:
                return_items.append(flip_lr)
                return_items.append(flip_ud)
        else:
            if self.random_flip:
                return_items.append(psf_aug)
                return_items.append(flip_lr)
                return_items.append(flip_ud)

        # Add background to achieve desired SNR
        if self.bg_sim is not None:
            sig_var = torch.var(lensless.flatten())
            target_snr = np.random.uniform(self.bg_snr_range[0], self.bg_snr_range[1])
            alpha = torch.sqrt(sig_var / self.background_var / (10**target_snr / 10))

            scaled_bg = alpha * self.bg_sim

            # Add background noise to the target image
            image_with_bg = lensless + scaled_bg

            # Add simulated background to get image_with_bg and scaled_bg
            return_items[0] = image_with_bg
            return_items[0].append(scaled_bg)
        # If measured background available in the dataset return it
        elif self.measured_bg:
            return_items.append(background)
        return return_items

    def extract_roi(
        self,
        reconstruction,
        lensed=None,
        axis=(1, 2),
        flip_lr=None,
        flip_ud=None,
        rotate_aug=False,
        shift_aug=None,
    ):
        """
        Parameters
        ----------
        flip_lr : torch.Tensor, optional
            Tensor of booleans indicating whether to flip the reconstruction left-right, by default None.
        flip_ud : bool, optional
            Tensor of booleans indicating whether to flip the reconstruction up-down, by default None.
        """
        n_dim = len(reconstruction.shape)
        assert max(axis) < n_dim, "Axis should be within the dimensions of the reconstruction."

        # add batch dimension
        if n_dim == 3:
            if isinstance(reconstruction, torch.Tensor):
                reconstruction = reconstruction.unsqueeze(0)
                if lensed is not None:
                    lensed = lensed.unsqueeze(0)
            else:
                reconstruction = reconstruction[np.newaxis]
                if lensed is not None:
                    lensed = lensed[np.newaxis]
            # increment axis
            axis = (axis[0] + 1, axis[1] + 1)

        # flip/rotate before alignment, as alignment parameters are assuming no flip/rotate
        if flip_lr is not None:
            flip_lr = flip_lr.squeeze().tolist()
            if isinstance(reconstruction, torch.Tensor):
                reconstruction[flip_lr] = torch.flip(reconstruction[flip_lr], dims=(axis[1],))
                if lensed is not None:
                    lensed[flip_lr] = torch.flip(lensed[flip_lr], dims=(axis[1],))
            else:
                reconstruction[flip_lr] = np.flip(reconstruction[flip_lr], axis=axis[1])
                if lensed is not None:
                    lensed[flip_lr] = np.flip(lensed[flip_lr], axis=axis[1])
        if flip_ud is not None:
            flip_ud = flip_ud.squeeze().tolist()
            if isinstance(reconstruction, torch.Tensor):
                reconstruction[flip_ud] = torch.flip(reconstruction[flip_ud], dims=(axis[0],))
                if lensed is not None:
                    lensed[flip_ud] = torch.flip(lensed[flip_ud], dims=(axis[0],))
            else:
                reconstruction[flip_ud] = np.flip(reconstruction[flip_ud], axis=axis[0])
                if lensed is not None:
                    lensed[flip_ud] = np.flip(lensed[flip_ud], axis=axis[0])
        if rotate_aug:
            assert isinstance(rotate_aug, float)
            if isinstance(reconstruction, torch.Tensor):
                assert axis == (-2, -1), "Only ...HW rotation is supported for torch.Tensor"
                reconstruction = F.rotate(reconstruction, -rotate_aug, expand=False)
                if lensed is not None:
                    lensed = F.rotate(lensed, -rotate_aug, expand=False)
            else:
                reconstruction = rotate(reconstruction, angle=-rotate_aug, axes=axis, reshape=False)
                if lensed is not None:
                    lensed = rotate(lensed, angle=-rotate_aug, axes=axis, reshape=False)
        if shift_aug is not None:
            assert isinstance(shift_aug, tuple)
            neg_shift = (-shift_aug[0], -shift_aug[1])
            reconstruction = shift_with_pad(reconstruction, neg_shift, axis=axis)
            if lensed is not None:
                lensed = shift_with_pad(lensed, neg_shift, axis=axis)

        if self.alignment is not None:
            top_left = self.alignment["top_left"]
            height = self.alignment["height"]
            width = self.alignment["width"]

            # extract according to axis
            index = [slice(None)] * n_dim
            index[axis[0]] = slice(top_left[0], top_left[0] + height)
            index[axis[1]] = slice(top_left[1], top_left[1] + width)
            reconstruction = reconstruction[tuple(index)]

            # rotate if necessary
            angle = self.alignment.get("angle", 0)
            if isinstance(reconstruction, torch.Tensor) and angle:
                reconstruction = F.rotate(reconstruction, angle, expand=False)
            elif angle:
                reconstruction = rotate(reconstruction, angle, axes=axis, reshape=False)

        elif self.crop is not None:
            vertical = self.crop["vertical"]
            horizontal = self.crop["horizontal"]

            # extract according to axis
            index = [slice(None)] * n_dim
            index[axis[0]] = slice(vertical[0], vertical[1])
            index[axis[1]] = slice(horizontal[0], horizontal[1])
            reconstruction = reconstruction[tuple(index)]
            if lensed is not None:
                lensed = lensed[tuple(index)]

        # flip/rotate back
        if flip_lr is not None:
            if isinstance(reconstruction, torch.Tensor):
                reconstruction[flip_lr] = torch.flip(reconstruction[flip_lr], dims=(axis[1],))
                if lensed is not None:
                    lensed[flip_lr] = torch.flip(lensed[flip_lr], dims=(axis[1],))
            else:
                reconstruction[flip_lr] = np.flip(reconstruction[flip_lr], axis=axis[1])
                if lensed is not None:
                    lensed[flip_lr] = np.flip(lensed[flip_lr], axis=axis[1])
        if flip_ud is not None:
            if isinstance(reconstruction, torch.Tensor):
                reconstruction[flip_ud] = torch.flip(reconstruction[flip_ud], dims=(axis[0],))
                if lensed is not None:
                    lensed[flip_ud] = torch.flip(lensed[flip_ud], dims=(axis[0],))
            else:
                reconstruction[flip_ud] = np.flip(reconstruction[flip_ud], axis=axis[0])
                if lensed is not None:
                    lensed[flip_ud] = np.flip(lensed[flip_ud], axis=axis[0])
        if rotate_aug:
            if isinstance(reconstruction, torch.Tensor):
                assert axis == (-2, -1), "Only ...HW rotation is supported for torch.Tensor"
                reconstruction = F.rotate(reconstruction, rotate_aug, expand=False)
                if lensed is not None:
                    lensed = F.rotate(lensed, rotate_aug, expand=False)
            else:
                reconstruction = rotate(reconstruction, angle=rotate_aug, axes=axis, reshape=False)
                if lensed is not None:
                    lensed = rotate(lensed, angle=rotate_aug, axes=axis, reshape=False)

        if shift_aug is not None:
            reconstruction = shift_with_pad(reconstruction, shift_aug, axis=axis)
            if lensed is not None:
                lensed = shift_with_pad(lensed, shift_aug, axis=axis)

        # remove batch dimension
        if n_dim == 3:
            if isinstance(reconstruction, torch.Tensor):
                reconstruction = reconstruction.squeeze(0)
                if lensed is not None:
                    lensed = lensed.squeeze(0)
            else:
                reconstruction = reconstruction[0]
                if lensed is not None:
                    lensed = lensed[0]

        if self.alignment is None and lensed is not None:
            return reconstruction, lensed
        else:
            return reconstruction


def simulate_dataset(config, generator=None):
    """
    Prepare datasets for training and testing.

    Parameters
    ----------
    config : omegaconf.DictConfig
        Configuration, e.g. from Hydra. See ``scripts/recon/train_learning_based.py`` for an example that uses this function.
    generator : torch.Generator, optional
        Random number generator, by default ``None``.
    """

    if "cuda" in config.torch_device and torch.cuda.is_available():
        device = config.torch_device
    else:
        device = "cpu"

    # -- prepare PSF
    psf = None
    if config.trainable_mask.mask_type is None or config.trainable_mask.initial_value == "psf":
        psf_fp = os.path.join(get_original_cwd(), config.files.psf)
        psf, _ = load_psf(
            psf_fp,
            downsample=config.files.downsample,
            return_float=True,
            return_bg=True,
            bg_pix=(0, 15),
        )
        if config.files.diffusercam_psf:
            transform_BRG2RGB = transforms.Lambda(lambda x: x[..., [2, 1, 0]])
            psf = transform_BRG2RGB(torch.from_numpy(psf))

        # drop depth dimension
        psf = psf.to(device)

    else:
        # training mask / PSF
        mask = prep_trainable_mask(config, psf)
        psf = mask.get_psf().to(device)

    # -- load dataset
    pre_transform = None
    transforms_list = [transforms.ToTensor()]
    data_path = os.path.join(get_original_cwd(), "data")
    if config.simulation.grayscale:
        transforms_list.append(transforms.Grayscale())

    if config.files.dataset == "mnist":
        transform = transforms.Compose(transforms_list)
        train_ds = datasets.MNIST(root=data_path, train=True, download=True, transform=transform)
        test_ds = datasets.MNIST(root=data_path, train=False, download=True, transform=transform)

    elif config.files.dataset == "fashion_mnist":
        transform = transforms.Compose(transforms_list)
        train_ds = datasets.FashionMNIST(
            root=data_path, train=True, download=True, transform=transform
        )
        test_ds = datasets.FashionMNIST(
            root=data_path, train=False, download=True, transform=transform
        )
    elif config.files.dataset == "cifar10":
        transform = transforms.Compose(transforms_list)
        train_ds = datasets.CIFAR10(root=data_path, train=True, download=True, transform=transform)
        test_ds = datasets.CIFAR10(root=data_path, train=False, download=True, transform=transform)

    elif config.files.dataset == "CelebA":
        root = config.files.celeba_root
        data_path = os.path.join(root, "celeba")
        assert os.path.isdir(
            data_path
        ), f"Data path {data_path} does not exist. Make sure you download the CelebA dataset and provide the parent directory as 'config.files.celeba_root'. Download link: https://mmlab.ie.cuhk.edu.hk/projects/CelebA.html"
        transform = transforms.Compose(transforms_list)
        if config.files.n_files is None:
            train_ds = datasets.CelebA(
                root=root, split="train", download=False, transform=transform
            )
            test_ds = datasets.CelebA(root=root, split="test", download=False, transform=transform)
        else:
            ds = datasets.CelebA(root=root, split="all", download=False, transform=transform)

            ds = Subset(ds, np.arange(config.files.n_files))

            train_size = int((1 - config.files.test_size) * len(ds))
            test_size = len(ds) - train_size
            train_ds, test_ds = torch.utils.data.random_split(
                ds, [train_size, test_size], generator=generator
            )
    else:
        raise NotImplementedError(f"Dataset {config.files.dataset} not implemented.")

    if config.files.dataset != "CelebA":
        if config.files.n_files is not None:
            train_size = int((1 - config.files.test_size) * config.files.n_files)
            test_size = config.files.n_files - train_size
            train_ds = Subset(train_ds, np.arange(train_size))
            test_ds = Subset(test_ds, np.arange(test_size))

    # convert PSF
    if config.simulation.grayscale and not is_grayscale(psf):
        psf = rgb2gray(psf)

    # check if gpu is available
    device_conv = config.torch_device
    if device_conv == "cuda" and torch.cuda.is_available():
        device_conv = "cuda"
    else:
        device_conv = "cpu"

    # create simulator
    simulator = FarFieldSimulator(
        psf=psf,
        is_torch=True,
        **config.simulation,
    )

    # create Pytorch dataset and dataloader
    crop = config.files.crop.copy() if config.files.crop is not None else None
    if mask is None:
        train_ds_prop = SimulatedFarFieldDataset(
            dataset=train_ds,
            simulator=simulator,
            dataset_is_CHW=True,
            device_conv=device_conv,
            flip=config.simulation.flip,
            vertical_shift=config.files.vertical_shift,
            horizontal_shift=config.files.horizontal_shift,
            crop=crop,
            downsample=config.files.downsample,
            pre_transform=pre_transform,
        )
        test_ds_prop = SimulatedFarFieldDataset(
            dataset=test_ds,
            simulator=simulator,
            dataset_is_CHW=True,
            device_conv=device_conv,
            flip=config.simulation.flip,
            vertical_shift=config.files.vertical_shift,
            horizontal_shift=config.files.horizontal_shift,
            crop=crop,
            downsample=config.files.downsample,
            pre_transform=pre_transform,
        )
    else:
        if config.measure is not None:

            train_ds_prop = HITLDatasetTrainableMask(
                rpi_username=config.measure.rpi_username,
                rpi_hostname=config.measure.rpi_hostname,
                celeba_root=config.files.celeba_root,
                display_config=config.measure.display,
                capture_config=config.measure.capture,
                mask_center=config.trainable_mask.ap_center,
                dataset=train_ds,
                mask=mask,
                simulator=simulator,
                dataset_is_CHW=True,
                device_conv=device_conv,
                flip=config.simulation.flip,
                vertical_shift=config.files.vertical_shift,
                horizontal_shift=config.files.horizontal_shift,
                crop=crop,
                downsample=config.files.downsample,
                pre_transform=pre_transform,
            )

            test_ds_prop = HITLDatasetTrainableMask(
                rpi_username=config.measure.rpi_username,
                rpi_hostname=config.measure.rpi_hostname,
                celeba_root=config.files.celeba_root,
                display_config=config.measure.display,
                capture_config=config.measure.capture,
                mask_center=config.trainable_mask.ap_center,
                dataset=test_ds,
                mask=mask,
                simulator=simulator,
                dataset_is_CHW=True,
                device_conv=device_conv,
                flip=config.simulation.flip,
                vertical_shift=config.files.vertical_shift,
                horizontal_shift=config.files.horizontal_shift,
                crop=crop,
                downsample=config.files.downsample,
                pre_transform=pre_transform,
            )

        else:

            train_ds_prop = SimulatedDatasetTrainableMask(
                dataset=train_ds,
                mask=mask,
                simulator=simulator,
                dataset_is_CHW=True,
                device_conv=device_conv,
                flip=config.simulation.flip,
                vertical_shift=config.files.vertical_shift,
                horizontal_shift=config.files.horizontal_shift,
                crop=crop,
                downsample=config.files.downsample,
                pre_transform=pre_transform,
            )
            test_ds_prop = SimulatedDatasetTrainableMask(
                dataset=test_ds,
                mask=mask,
                simulator=simulator,
                dataset_is_CHW=True,
                device_conv=device_conv,
                flip=config.simulation.flip,
                vertical_shift=config.files.vertical_shift,
                horizontal_shift=config.files.horizontal_shift,
                crop=crop,
                downsample=config.files.downsample,
                pre_transform=pre_transform,
            )

    return train_ds_prop, test_ds_prop, mask


class MyDataParallel(torch.nn.DataParallel):
    def __getattr__(self, name):
        try:
            return super().__getattr__(name)
        except AttributeError:
            return getattr(self.module, name)



# #############################################################################
# integrated_background_sub.py
# ============================
# Authors :
# Eric BEZZAM [ebezzam@gmail.com]
# Stefan PETERS
# #############################################################################


import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
from lensless.recon.utils import convert_to_NCHW, convert_to_NDCHW


class DoubleConv(nn.Module):
    """(convolution => [BN] => ReLU) * 2"""

    def __init__(self, in_channels, out_channels, mid_channels=None):
        super().__init__()
        if not mid_channels:
            mid_channels = out_channels
        self.double_conv = nn.Sequential(
            nn.Conv2d(in_channels, mid_channels, kernel_size=(3, 3), padding=1, bias=False),
            nn.BatchNorm2d(mid_channels),
            nn.ReLU(inplace=True),
            nn.Conv2d(mid_channels, out_channels, kernel_size=(3, 3), padding=1, bias=False),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True),
        )

    def forward(self, x):
        return self.double_conv(x)


class Down(nn.Module):
    def __init__(self, in_channels, out_channels):
        super().__init__()
        self.pool_conv = nn.Sequential(
            nn.MaxPool2d(2),
            DoubleConv(in_channels, out_channels),
        )

    def forward(self, x):
        return self.pool_conv(x)


class Down_concat(nn.Module):
    def __init__(self, in_channels, out_channels):
        super().__init__()

        # self.down = nn.Downsample(scale_factor=2, mode="bilinear", align_corners=True)
        # self.conv = DoubleConv(in_channels, out_channels, in_channels // 2)

        self.pool_conv = nn.Sequential(
            nn.MaxPool2d(2),
            DoubleConv(in_channels * 2, out_channels),
        )

    def forward(self, x1, x2):
        # # don't need passing as should be same shape...
        # diffY = x2.size()[2] - x1.size()[2]
        # diffX = x2.size()[3] - x1.size()[3]
        # x1 = F.pad(x1, [diffX // 2, diffX - diffX // 2, diffY // 2, diffY - diffY // 2])

        x = torch.cat([x1, x2], dim=1)
        return self.pool_conv(x)


class Up(nn.Module):
    def __init__(self, in_channels, out_channels):
        super().__init__()

        # or use ConvTranspose2d? https://github.com/milesial/Pytorch-UNet/blob/21d7850f2af30a9695bbeea75f3136aa538cfc4a/unet/unet_parts.py#L53
        self.up = nn.Upsample(scale_factor=2, mode="bilinear", align_corners=True)
        self.conv = DoubleConv(in_channels, out_channels, in_channels // 2)

    def forward(self, x1, x2):
        x1 = self.up(x1)
        # input is CHW
        diffY = x2.size()[2] - x1.size()[2]
        diffX = x2.size()[3] - x1.size()[3]

        x1 = F.pad(x1, [diffX // 2, diffX - diffX // 2, diffY // 2, diffY - diffY // 2])
        x = torch.cat([x2, x1], dim=1)
        return self.conv(x)


class OutConv(nn.Module):
    def __init__(self, in_channels, out_channels):
        super(OutConv, self).__init__()
        self.conv = nn.Conv2d(in_channels, out_channels, kernel_size=(1, 1))

    def forward(self, x):
        return self.conv(x)


class IntegratedBackgroundSub(nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        input_shape,
        down_subtraction=False,  # TODO flag to do background subtraction while downsampling (instead of upsampling)
        nc=None,
    ):
        """
        Integrated Background Subtraction using a U-Net architecture.

        Parameters
        ----------
        in_channels : int TODO
            Number of input channels. RGB or grayscale, i.e. 3 and 1 respectively.
        out_channels : int
            Number of output channels. RGB or grayscale, i.e. 3 and 1 respectively.
        psf : :py:class:`~torch.Tensor`
            Point spread function (PSF) that models forward propagation.
        psf_channels : int
            Number of channels in the PSF. Default is 1.
        down_subtraction : bool
            Flag to do background subtraction while downsampling (instead of upsampling).
        nc : list
            Number of channels in the network. Default is [64, 128, 256, 512, 512].

        """
        assert in_channels == 1 or in_channels == 3, "in_channels must be 1 or 3"
        assert out_channels == 1 or out_channels == 3, "out_channels must be 1 or 3"
        assert in_channels >= out_channels
        if nc is None:
            nc = [64, 128, 256, 512, 512]

        super(IntegratedBackgroundSub, self).__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels

        self.inc = DoubleConv(in_channels, nc[0])
        self.down_subtraction = down_subtraction

        if self.down_subtraction:
            self.down_layers = nn.ModuleList(
                [Down_concat(nc[i], nc[i + 1]) for i in range(len(nc) - 1)]
            )
        else:
            self.down_layers = nn.ModuleList([Down(nc[i], nc[i + 1]) for i in range(len(nc) - 1)])

        # background downsampling layers
        self.inc_back = DoubleConv(in_channels, nc[0])
        self.down_layers_bg = nn.ModuleList([Down(nc[i], nc[i + 1]) for i in range(len(nc) - 2)])
        self.subtraction_weight = nn.Parameter(torch.ones(len(nc) - 1))

        # upsampling layers
        self.up_layers = []
        n_prev = nc[-1]
        for i in range(len(nc) - 1):
            n_in = n_prev + nc[-i - 2]
            n_out = nc[-i - 2] // 2 if i < len(nc) - 2 else nc[0]
            self.up_layers.append(Up(n_in, n_out))
            n_prev = n_out
        self.up_layers = nn.ModuleList(self.up_layers)
        self.outc = OutConv(nc[0], out_channels)

        # padding H and W to next multiple of 8
        self.top = (8 - input_shape[0] % 8) // 2
        self.bottom = (8 - input_shape[0] % 8) - self.top
        self.left = (8 - input_shape[1] % 8) // 2
        self.right = (8 - input_shape[1] % 8) - self.left

    def forward(self, batch, background=None, **kwargs):
        assert batch.shape[-3:-1] == background.shape[-3:-1], "Data and background shape mismatch"
        assert len(batch.shape) >= 3, "Batch must have at least 3 dimensions"

        n_depth = batch[0].shape[-4]
        if n_depth > 1:
            raise NotImplementedError("3D not implemented yet.")

        # pad to multiple of 8
        batch = convert_to_NCHW(batch)
        batch = torch.nn.functional.pad(
            batch, (self.left, self.right, self.top, self.bottom), mode="constant", value=0
        )
        background = convert_to_NCHW(background)
        background = torch.nn.functional.pad(
            background, (self.left, self.right, self.top, self.bottom), mode="constant", value=0
        )

        # Select the images
        x_inter = [self.inc(batch)]

        # -- downscale background
        bg_inter = [self.inc_back(background)]
        for i in range(len(self.down_layers_bg)):
            bg_inter.append(self.down_layers_bg[i](bg_inter[-1]))

        if self.down_subtraction:
            # Concatenate background subtracted data during downsampling
            batch = x_inter[0]

            # -- downsample
            # for i in range(len(bg_inter)):
            #    x_inter.append(self.down_layers[i](batch - bg_inter[i], bg_inter[i]))
            # [(M-B), B]
            for i in range(len(bg_inter)):
                # -- [(M-B), M]
                # batch = self.down_layers[i](batch - self.subtraction_weight[i] * bg_inter[i], batch)
                # -- [(M-B), B]
                batch = self.down_layers[i](
                    batch - self.subtraction_weight[i] * bg_inter[i], bg_inter[i]
                )
                x_inter.append(batch)

            # upsample
            batch = self.up_layers[0](x_inter[-1], x_inter[-2])
            for i in range(len(self.up_layers) - 1):
                batch = self.up_layers[i + 1](batch, x_inter[-i - 3])

        else:
            # Concatenate background subtracted data during upsampling
            # -- downsample
            for i in range(len(self.down_layers)):
                x_inter.append(self.down_layers[i](x_inter[-1]))

            # simple subtraction between the downsapled backgrounds and the corresponding data
            for i in range(len(bg_inter)):
                x_inter[i] = x_inter[i] - self.subtraction_weight[i] * bg_inter[i]

            # upsample
            batch = self.up_layers[0](x_inter[-1], x_inter[-2])
            for i in range(len(self.up_layers) - 1):
                batch = self.up_layers[i + 1](batch, x_inter[-i - 3])

        batch = self.outc(batch)

        # back to original shape
        batch = batch[..., self.top : -self.bottom, self.left : -self.right]
        batch = convert_to_NDCHW(batch, n_depth)

        # normalize to [0,1], TODO use sigmoid instead?
        batch = (batch + 1) / 2
        batch = torch.clip(batch, min=0.0)

        return batch



"""
Push DiffuserCam dataset to HuggingFace.

```bash
# install
pip install datasets
pip install huggingface_hub
pip install joblib
pip install git+https://github.com/pvigier/perlin-numpy.git@5e26837db14042e51166eb6cad4c0df2c1907016
pip install git+https://github.com/ebezzam/slm-controller.git

# make a write token on HuggingFace

# run
python scripts/data/upload_diffusercam_huggingface.py \
hf_token=... \
```
"""

import hydra
import time
import numpy as np
import os
import glob
from lensless.utils.io import save_image
import cv2
from datasets import Dataset, DatasetDict, Image
from huggingface_hub import upload_file
from lensless.utils.dataset import natural_sort
from tqdm import tqdm
from joblib import Parallel, delayed


@hydra.main(
    version_base=None,
    config_path="../../configs/dataset",
    config_name="upload_diffusercam_huggingface",
)
def upload_dataset(config):

    # parameters
    repo_id = config.repo_id
    dir_diffuser = config.dir_diffuser
    dir_lensed = config.dir_lensed
    psf_fp = config.psf_fp
    hf_token = config.hf_token
    file_ext = config.file_ext
    n_files = config.n_files
    n_jobs = config.n_jobs
    normalize = config.normalize

    assert hf_token is not None, "Please provide a HuggingFace token."

    start_time = time.time()

    # get all lensless-lensed pairs
    files_diffuser = glob.glob(os.path.join(dir_diffuser, "*" + file_ext))
    files_lensed = glob.glob(os.path.join(dir_lensed, "*" + file_ext))

    # only keep if in both
    bn_diffuser = [os.path.basename(f) for f in files_diffuser]
    bn_lensed = [os.path.basename(f) for f in files_lensed]
    common_files = list(set(bn_diffuser).intersection(bn_lensed))
    common_files = natural_sort(common_files)
    print(f"Number of common files: {len(common_files)}")
    if n_files is not None:
        print(f"Only keeping {n_files} files...")
        common_files = common_files[:n_files]

    # load PSF, convert to RGB, save as PNG
    # psf_img = np.array(PIL.Image.open(psf_fp))
    psf_img = cv2.imread(psf_fp, cv2.IMREAD_UNCHANGED)
    psf_img = cv2.cvtColor(psf_img, cv2.COLOR_BGR2RGB)  # convert to RGB
    psf_fp_png = psf_fp.replace(".tiff", ".png")
    save_image(psf_img, psf_fp_png, normalize=True)  # need normalize=True

    # save as PNG
    dir_diffuser_png = dir_diffuser.replace("diffuser_images", "diffuser_png")
    os.makedirs(dir_diffuser_png, exist_ok=True)
    dir_lensed_png = dir_lensed.replace("ground_truth_lensed", "lensed_png")
    os.makedirs(dir_lensed_png, exist_ok=True)

    # -- parallelize with joblib
    def save_png(f, dir_diffuser, dir_diffuser_png, dir_lensed, dir_lensed_png):

        diffuser_img = np.load(os.path.join(dir_diffuser, f))
        diffuser_img = cv2.cvtColor(diffuser_img, cv2.COLOR_BGR2RGB)  # convert to RGB
        diffuser_fn = os.path.join(dir_diffuser_png, f.replace(file_ext, ".png"))
        save_image(diffuser_img, diffuser_fn, normalize=normalize)

        lensed_img = np.load(os.path.join(dir_lensed, f))
        lensed_img = cv2.cvtColor(lensed_img, cv2.COLOR_BGR2RGB)  # convert to RGB
        lensed_fn = os.path.join(dir_lensed_png, f.replace(file_ext, ".png"))
        save_image(lensed_img, lensed_fn, normalize=normalize)

    Parallel(n_jobs=n_jobs)(
        delayed(save_png)(f, dir_diffuser, dir_diffuser_png, dir_lensed, dir_lensed_png)
        for f in tqdm(common_files)
    )

    # get file paths
    diffuser_files = [
        os.path.join(dir_diffuser_png, f.replace(file_ext, ".png")) for f in common_files
    ]
    lensed_files = [os.path.join(dir_lensed_png, f.replace(file_ext, ".png")) for f in common_files]
    diffuser_files = natural_sort(diffuser_files)
    lensed_files = natural_sort(lensed_files)

    # step 1: create Dataset objects
    def create_dataset(diffuser_files, lensed_files):
        dataset = Dataset.from_dict(
            {
                "lensless": diffuser_files,
                "lensed": lensed_files,
            }
        )
        dataset = dataset.cast_column("lensless", Image())
        dataset = dataset.cast_column("lensed", Image())
        return dataset

    # according to original split test files are up to idx=1000, for some reason im1 is missing?
    test_dataset = create_dataset(diffuser_files[:999], lensed_files[:999])
    train_dataset = create_dataset(diffuser_files[999:], lensed_files[999:])

    # step 2: create DatasetDict
    dataset_dict = DatasetDict(
        {
            "train": train_dataset,
            "test": test_dataset,
        }
    )

    # step 3: push to hub
    upload_file(
        path_or_fileobj=psf_fp,
        path_in_repo="psf.tiff",
        repo_id=repo_id,
        repo_type="dataset",
        token=hf_token,
    )

    # -- dataset
    dataset_dict.push_to_hub(
        repo_id,
        token=hf_token,
    )
    upload_file(
        path_or_fileobj=psf_fp_png,
        path_in_repo="psf.png",
        repo_id=repo_id,
        repo_type="dataset",
        token=hf_token,
    )
    upload_file(
        path_or_fileobj=diffuser_files[0],
        path_in_repo="lensless_example.png",
        repo_id=repo_id,
        repo_type="dataset",
        token=hf_token,
    )
    upload_file(
        path_or_fileobj=lensed_files[0],
        path_in_repo="lensed_example.png",
        repo_id=repo_id,
        repo_type="dataset",
        token=hf_token,
    )

    # delete PNG files
    os.system(f"rm -rf {dir_diffuser_png}")
    os.system(f"rm -rf {dir_lensed_png}")

    # total time in minutes
    print(f"Total time: {(time.time() - start_time) / 60} minutes")


if __name__ == "__main__":
    upload_dataset()



# #############################################################################
# unrolled_fista.py
# =================
# Authors :
# Yohann PERRON [yohann.perron@gmail.com]
# #############################################################################

import numpy as np
from lensless.recon.trainable_recon import TrainableReconstructionAlgorithm
from lensless.recon.gd import non_neg

try:
    import torch

except ImportError:
    raise ImportError("Pytorch is require to use trainable reconstruction algorithms.")


class UnrolledFISTA(TrainableReconstructionAlgorithm):
    """
    Object for applying unrolled projected gradient descent with FISTA (Fast Iterative
    Shrinkage-Thresholding Algorithm) for acceleration.

    FISTA Paper: https://www.ceremade.dauphine.fr/~carlier/FISTA

    """

    def __init__(self, psf, n_iter=5, dtype=None, proj=non_neg, learn_tk=True, tk=1, **kwargs):
        """
        COnstructor for unrolled FISTA algorithm.

        Parameters
        ----------
        psf : :py:class:`~torch.Tensor`
            Point spread function (PSF) that models forward propagation.
            Must be of shape (depth, height, width, channels) even if
            depth = 1 and channels = 1. You can use :py:func:`~lensless.io.load_psf`
            to load a PSF from a file such that it is in the correct format.
        n_iter : int, optional
            Number of iterations to unrolled, by default 5
        dtype : float32 or float64
            Data type to use for optimization.
        proj : :py:class:`function`, optional
            Projection function to apply at each iteration, by default non_neg
        learn_tk : bool, optional
            whether the tk parameters of FISTA should be learnt, by default True
        tk : int, optional
            Initial value of tk, by default 1
        """

        super(UnrolledFISTA, self).__init__(psf, n_iter=n_iter, dtype=dtype, reset=False, **kwargs)

        self._proj = proj

        # initial guess, half intensity image
        # for online approach could use last reconstruction
        psf_flat = self._psf.reshape(-1, self._psf_shape[3])
        pixel_start = (torch.max(psf_flat, axis=0).values + torch.min(psf_flat, axis=0).values) / 2
        self._image_init = torch.ones_like(self._psf[None, ...]) * pixel_start

        # learnable step size initialize as < 2 / lipschitz
        Hadj_flat = self._convolver._Hadj.reshape(-1, self._psf_shape[3])
        H_flat = self._convolver._H.reshape(-1, self._psf_shape[3])
        if not self.skip_unrolled:
            self._alpha_p = torch.nn.Parameter(
                torch.ones(self._n_iter, self._psf_shape[3]).to(psf.device)
                * (1.8 / torch.max(torch.abs(Hadj_flat * H_flat), axis=0).values)
            )
        else:
            self._alpha_p = torch.ones(self._n_iter, self._psf_shape[3]).to(psf.device) * (
                1.8 / torch.max(torch.abs(Hadj_flat * H_flat), axis=0).values
            )

        # set tk, can be learnt if learn_tk=True
        self._tk_p = [tk]
        for i in range(self._n_iter):
            self._tk_p.append((1 + np.sqrt(1 + 4 * self._tk_p[i] ** 2)) / 2)
        self._tk_p = torch.Tensor(self._tk_p)
        if learn_tk and not self.skip_unrolled:
            self._tk_p = torch.nn.Parameter(self._tk_p).to(psf.device)

        self.reset()

    def _form_image(self):
        return self._proj(self._image_est)

    def _grad(self):
        diff = self._convolver.convolve(self._image_est) - self._data
        return self._convolver.deconvolve(diff)

    def reset(self, batch_size=1):
        if self._initial_est is not None:
            self._image_est = self._initial_est
        else:
            self._image_est = self._image_init.expand(batch_size, -1, -1, -1, -1)
        self._xk = self._image_est

        # enforce positivity
        self._alpha = torch.abs(self._alpha_p)
        self._tk = torch.abs(self._tk_p)

    def _update(self, iter):
        self._image_est = self._image_est - self._alpha[iter] * self._grad()
        xk = self._proj(self._image_est)
        self._image_est = xk + (self._tk[iter] - 1) / self._tk[iter + 1] * (xk - self._xk)
        self._xk = xk



# #############################################################################
# benchmark.py
# =================
# Authors :
# Yohann PERRON
# Eric BEZZAM [ebezzam@gmail.com]
# #############################################################################


from lensless.utils.dataset import DiffuserCamTestDataset
from lensless.utils.io import save_image
from waveprop.noise import add_shot_noise
from tqdm import tqdm
import os
import numpy as np
import wandb

try:
    import torch
    from torch.utils.data import DataLoader
    from torch.nn import MSELoss, L1Loss
    from torchmetrics import StructuralSimilarityIndexMeasure
    from torchmetrics.image import lpip, psnr
    from torch.optim import SGD
except ImportError:
    raise ImportError(
        "Torch, torchvision, and torchmetrics are needed to benchmark reconstruction algorithm."
    )


# TODO put Parameterize and Perturb functions in a separate file
def get_param(model):
    all_param = []
    for _, param in model.named_parameters():
        all_param.append(param)
    all_param = torch.cat([param.flatten() for param in all_param])
    return all_param


def pnp_loss(y_est, y, recon, mu, original_param):
    new_param = get_param(recon)
    loss = torch.mean((y_est - y) ** 2) + mu * torch.mean((new_param - original_param) ** 2)
    return loss


def swap_color_channels(tensor, channels):
    if len(channels) == 2:
        color1 = tensor[..., channels[0]].copy()
        tensor[..., channels[0]] = tensor[..., channels[1]]
        tensor[..., channels[1]] = color1
    elif len(channels) == 3:
        # reorder channels
        color1 = tensor[..., channels[0]].copy()
        color2 = tensor[..., channels[1]].copy()
        color3 = tensor[..., channels[2]].copy()
        tensor[..., 0] = color1
        tensor[..., 1] = color2
        tensor[..., 2] = color3
    return tensor


def benchmark(
    model,
    dataset,
    batchsize=1,
    metrics=None,
    crop=None,
    save_idx=None,
    save_intermediate=False,
    output_dir=None,
    unrolled_output_factor=False,
    pre_process_aux=False,
    return_average=True,
    snr=None,
    use_wandb=False,
    label=None,
    epoch=None,
    use_background=True,
    pnp=None,
    swap_channels=False,
    **kwargs,
):
    """
    Compute multiple metrics for a reconstruction algorithm.

    Parameters
    ----------
    model : :py:class:`~lensless.ReconstructionAlgorithm`
        Reconstruction algorithm to benchmark.
    dataset : :py:class:`~lensless.benchmark.ParallelDataset`
        Parallel dataset of lensless and lensed images.
    batchsize : int, optional
        Batch size for processing. For maximum compatibility use 1 (batchsize above 1 are not supported on all algorithm), by default 1
    metrics : dict, optional
        Dictionary of metrics to compute. If None, MSE, MAE, SSIM, LPIPS and PSNR are computed.
    save_idx : list of int, optional
        List of indices to save the predictions, by default None (not to save any).
    output_dir : str, optional
        Directory to save the predictions, by default save in working directory if save_idx is provided.
    crop : dict, optional
        Dictionary of crop parameters (vertical: [start, end], horizontal: [start, end]), by default None (no crop).
    unrolled_output_factor : bool, optional
        If True, compute metrics for unrolled output, by default False.
    return_average : bool, optional
        If True, return the average value of the metrics, by default True.
    snr : float, optional
        Signal to noise ratio for adding shot noise. If None, no noise is added, by default None.
    use_background: bool, optional
        If dataset has background, use it for reconstruction, by default True.
    pnp : dict, optional
        Dictionary of parameters for (Parameterize and perturb) algorithm, by default None.
        Required keys: "mu" (distance from original parameters), "lr" (SGD learning rate), "n_iter" (number of iterations), "model_path" (original model path).

    Returns
    -------
    Dict[str, float]
        A dictionary containing the metrics name and average value
    """
    assert isinstance(model._psf, torch.Tensor), "model need to be constructed with torch support"
    device = model._psf.device

    if output_dir is None:
        output_dir = os.getcwd()
    else:
        output_dir = str(output_dir)
        if not os.path.exists(output_dir):
            os.mkdir(output_dir)

    if pnp is not None:
        assert "mu" in pnp, "mu must be provided"
        assert "lr" in pnp, "lr must be provided"
        assert "n_iter" in pnp, "n_iter must be provided"
        assert "model_path" in pnp, "model_path must be provided"
        assert isinstance(pnp["mu"], float), "mu must be a float"
        assert isinstance(pnp["lr"], float), "lr must be a float"
        assert isinstance(pnp["n_iter"], int), "n_iter must be an int"
        assert isinstance(pnp["model_path"], str), "model_path must be a string"
        assert batchsize == 1, "batchsize must be 1 for parameterize and perturb"
        original_param = get_param(model)

    if metrics is None:
        metrics = {
            "MSE": MSELoss(reduction="mean").to(device),
            "LPIPS_Vgg": lpip.LearnedPerceptualImagePatchSimilarity(
                net_type="vgg", normalize=True, reduction="sum"
            ).to(device),
            # "LPIPS_Alex": lpip.LearnedPerceptualImagePatchSimilarity(
            #     net_type="alex", normalize=True
            # ).to(device),
            "PSNR": psnr.PeakSignalNoiseRatio(reduction=None, dim=(1, 2, 3), data_range=(0, 1)).to(
                device
            ),
            "SSIM": StructuralSimilarityIndexMeasure(reduction=None, data_range=(0, 1)).to(device),
            "ReconstructionError": None,
        }

    metrics_values = {key: [] for key in metrics}
    if unrolled_output_factor:
        output_metrics = metrics.keys()
        for key in output_metrics:
            if key != "ReconstructionError":
                metrics_values[key + "_unrolled"] = []
    if pre_process_aux:
        metrics_values["ReconstructionError_PreProc"] = []
    output_intermediate = unrolled_output_factor or pre_process_aux or save_intermediate

    # loop over batches
    dataloader = DataLoader(dataset, batch_size=batchsize, pin_memory=(device != "cpu"))
    model.reset()
    idx = 0
    weights = []  # for averaging batches
    for batch in tqdm(dataloader):
        weights.append(len(batch[0]))

        flip_lr = None
        flip_ud = None
        background = None
        lensless = batch[0].to(device)
        lensed = batch[1].to(device)
        if dataset.measured_bg and use_background:
            background = batch[-1].to(device)
        if dataset.multimask or dataset.random_flip:
            psfs = batch[2]
            psfs = psfs.to(device)
        else:
            psfs = None
        if dataset.random_flip:
            flip_lr = batch[3]
            flip_ud = batch[4]

        # add shot noise
        if snr is not None:
            for i in range(lensless.shape[0]):
                lensless[i] = add_shot_noise(lensless[i], float(snr))

        # compute predictions
        if batchsize == 1:

            if pnp is not None:
                from lensless.recon.model_dict import load_model
                from lensless.recon.rfft_convolve import RealFFTConvolve2D

                psf = psfs[0].to(device)
                recon_pnp = load_model(pnp["model_path"], psf, device=device, verbose=False)

                # define optimizer
                optimizer = SGD(recon_pnp.parameters(), lr=pnp["lr"])

                # create forward model (which needs padding)
                forward_model_param = recon_pnp._convolver_param.copy()
                forward_model_param["pad"] = True
                forward_model = RealFFTConvolve2D(psf, **forward_model_param)

                # iterate / minimize loss
                for i in range(pnp["n_iter"]):
                    optimizer.zero_grad()
                    recon_pnp.set_data(lensless)
                    prediction = recon_pnp.apply(
                        disp_iter=-1,
                        save=False,
                        gamma=None,
                        plot=False,
                    )
                    # simulate measurement
                    y_est = forward_model.convolve(prediction)
                    # -- normalize
                    y_est = y_est - y_est.min()
                    y_est = y_est / y_est.max()
                    loss = pnp_loss(y_est, lensless, recon_pnp, pnp["mu"], original_param)
                    loss.backward()
                    optimizer.step()
                    # if i % 10 == 0:
                    #     print(f"Loss {i}: {loss}")
                # prediction = recon_pnp.apply(
                #     disp_iter=-1,
                #     save=False,
                #     gamma=None,
                #     plot=False,
                # ).detach()
                prediction = prediction.detach()

            else:
                with torch.no_grad():
                    if psfs is not None:
                        model._set_psf(psfs[0])
                    model.set_data(lensless)
                    prediction = model.apply(
                        plot=False,
                        save=False,
                        output_intermediate=output_intermediate,
                        background=background,
                        **kwargs,
                    )

        else:
            with torch.no_grad():
                prediction = model.forward(
                    batch=lensless, psfs=psfs, background=background, **kwargs
                )

        if output_intermediate:
            psfs_out = prediction[3]
            pre_process_out = prediction[2]
            unrolled_out = prediction[1]
            prediction = prediction[0]
        prediction_original = prediction.clone()

        # Convert to [N*D, C, H, W] for torchmetrics
        prediction = prediction.reshape(-1, *prediction.shape[-3:]).movedim(-1, -3)
        lensed = lensed.reshape(-1, *lensed.shape[-3:]).movedim(-1, -3)

        if hasattr(dataset, "alignment"):
            if dataset.alignment is not None:
                prediction = dataset.extract_roi(
                    prediction, axis=(-2, -1), flip_lr=flip_lr, flip_ud=flip_ud
                )
            else:
                prediction, lensed = dataset.extract_roi(
                    prediction, axis=(-2, -1), lensed=lensed, flip_lr=flip_lr, flip_ud=flip_ud
                )
        elif crop is not None:
            assert flip_lr is None and flip_ud is None
            prediction = prediction[
                ...,
                crop["vertical"][0] : crop["vertical"][1],
                crop["horizontal"][0] : crop["horizontal"][1],
            ]
            lensed = lensed[
                ...,
                crop["vertical"][0] : crop["vertical"][1],
                crop["horizontal"][0] : crop["horizontal"][1],
            ]

        if save_idx is not None:
            for i, _batch_idx in enumerate(np.arange(idx, idx + batchsize)):
                if _batch_idx in save_idx:
                    prediction_np = prediction.cpu().numpy()[i]
                    # switch to [H, W, C] for saving
                    prediction_np = np.moveaxis(prediction_np, 0, -1)
                    fp = os.path.join(output_dir, f"{_batch_idx}.png")
                    save_image(prediction_np, fp=fp)

                    if save_intermediate:
                        fp = os.path.join(output_dir, f"{_batch_idx}_inv.png")
                        unrolled_out_np = unrolled_out.cpu().numpy()[i].squeeze()
                        # -- swap red and green channels
                        if swap_channels:
                            unrolled_out_np = swap_color_channels(unrolled_out_np, swap_channels)
                        save_image(unrolled_out_np, fp=fp)

                        fp = os.path.join(output_dir, f"{_batch_idx}_preproc.png")
                        pre_process_out_np = pre_process_out.cpu().numpy()[i].squeeze()
                        # -- swap red and green channels
                        if swap_channels:
                            pre_process_out_np = swap_color_channels(
                                pre_process_out_np, swap_channels
                            )
                        save_image(pre_process_out_np, fp=fp)

                        if psfs_out is not None:
                            fp = os.path.join(output_dir, f"{_batch_idx}_psfs.png")
                            if psfs_out.shape[0] == 1:
                                psfs_out_np = psfs_out.cpu().numpy().squeeze()
                            else:
                                psfs_out_np = psfs_out.cpu().numpy()[i].squeeze()

                            # -- swap red and green channels
                            if swap_channels:
                                psfs_out_np = swap_color_channels(psfs_out_np, swap_channels)
                            save_image(psfs_out_np, fp=fp)

                    if use_wandb:
                        assert epoch is not None, "epoch must be provided for wandb logging"
                        log_key = f"{_batch_idx}_{label}" if label is not None else f"{_batch_idx}"
                        wandb.log({log_key: wandb.Image(fp)}, step=epoch)

        # normalization
        prediction_max = torch.amax(prediction, dim=(-1, -2, -3), keepdim=True)
        if torch.all(prediction_max != 0):
            prediction = prediction / prediction_max
        else:
            print("Warning: prediction is zero")
        lensed_max = torch.amax(lensed, dim=(1, 2, 3), keepdim=True)
        lensed = lensed / lensed_max

        # compute metrics
        for metric in metrics:
            if metric == "ReconstructionError":
                metrics_values[metric] += model.reconstruction_error(
                    prediction=prediction_original, lensless=lensless, psfs=psfs
                ).tolist()
            else:
                try:
                    if "LPIPS" in metric:
                        if prediction.shape[1] == 1:
                            # LPIPS needs 3 channels
                            metrics_values[metric].append(
                                metrics[metric](
                                    prediction.repeat(1, 3, 1, 1), lensed.repeat(1, 3, 1, 1)
                                )
                                .cpu()
                                .item()
                            )
                        else:
                            metrics_values[metric].append(
                                metrics[metric](prediction, lensed).cpu().item()
                            )
                    elif metric == "MSE":
                        metrics_values[metric].append(
                            metrics[metric](prediction, lensed).cpu().item() * len(batch[0])
                        )
                    else:
                        vals = metrics[metric](prediction, lensed).cpu()
                        if hasattr(vals.tolist(), "__len__"):
                            metrics_values[metric] += vals.tolist()
                        else:
                            metrics_values[metric].append(vals.item())
                except Exception as e:
                    print(f"Error in metric {metric}: {e}")

        # compute metrics for unrolled output
        if unrolled_output_factor:

            # -- convert to CHW and remove depth
            unrolled_out = unrolled_out.reshape(-1, *unrolled_out.shape[-3:]).movedim(-1, -3)

            # -- extraction region of interest
            if hasattr(dataset, "alignment"):
                if dataset.alignment is not None:
                    unrolled_out = dataset.extract_roi(unrolled_out, axis=(-2, -1))
                else:
                    unrolled_out = dataset.extract_roi(
                        unrolled_out,
                        axis=(-2, -1),
                        # lensed=lensed   # lensed already extracted before
                    )
                assert np.all(lensed.shape == unrolled_out.shape)
            elif crop is not None:
                unrolled_out = unrolled_out[
                    ...,
                    crop["vertical"][0] : crop["vertical"][1],
                    crop["horizontal"][0] : crop["horizontal"][1],
                ]

            # -- normalization
            unrolled_out_max = torch.amax(unrolled_out, dim=(-1, -2, -3), keepdim=True)
            if torch.all(unrolled_out_max != 0):
                unrolled_out = unrolled_out / unrolled_out_max

            # -- compute metrics
            for metric in metrics:
                if metric == "ReconstructionError":
                    # only have this for final output
                    continue
                else:
                    if "LPIPS" in metric:
                        if unrolled_out.shape[1] == 1:
                            # LPIPS needs 3 channels
                            metrics_values[metric + "_unrolled"].append(
                                metrics[metric](
                                    unrolled_out.repeat(1, 3, 1, 1), lensed.repeat(1, 3, 1, 1)
                                )
                                .cpu()
                                .item()
                            )
                        else:
                            metrics_values[metric + "_unrolled"].append(
                                metrics[metric](unrolled_out, lensed).cpu().item()
                            )
                    elif metric == "MSE":
                        metrics_values[metric + "_unrolled"].append(
                            metrics[metric](unrolled_out, lensed).cpu().item() * len(batch[0])
                        )
                    else:
                        vals = metrics[metric](unrolled_out, lensed).cpu()
                        if hasattr(vals.tolist(), "__len__"):
                            metrics_values[metric + "_unrolled"] += vals.tolist()
                        else:
                            metrics_values[metric + "_unrolled"].append(vals.item())

        # compute metrics for pre-processed output
        if pre_process_aux:
            metrics_values["ReconstructionError_PreProc"] += model.reconstruction_error(
                prediction=prediction_original, lensless=pre_process_out
            ).tolist()

        model.reset()
        idx += batchsize

    # average metrics
    if return_average:
        for metric in metrics_values.keys():
            if "MSE" in metric or "LPIPS" in metric:
                # differently because metrics are grouped into bathces
                metrics_values[metric] = np.sum(metrics_values[metric]) / len(dataset)
            else:
                metrics_values[metric] = np.mean(metrics_values[metric])

    return metrics_values


if __name__ == "__main__":
    from lensless import ADMM

    downsample = 1.0
    batchsize = 1
    n_files = 10
    n_iter = 100

    # check if GPU is available
    if torch.cuda.is_available():
        device = "cuda"
    else:
        device = "cpu"

    # prepare dataset
    dataset = DiffuserCamTestDataset(n_files=n_files, downsample=downsample)

    # prepare model
    psf = dataset.psf.to(device)
    model = ADMM(psf, n_iter=n_iter)

    # run benchmark
    print(benchmark(model, dataset, batchsize=batchsize))



# #############################################################################
# tikhonov.py
# =================
# Authors :
# Eric BEZZAM [ebezzam@gmail.com]
# Aaron FARGEON [aa.fargeon@gmail.com]
# #############################################################################

"""
Tikhonov
========

The py:class:`~lensless.recon.tikhonov.CodedApertureReconstruction` class is meant
to recover an image from a py:class:`~lensless.hardware.mask.CodedAperture` lensless
capture, using the analytical solution to the Tikhonov optimization problem
(least squares problem with L2 regularization term), as in the `FlatCam paper
<https://arxiv.org/abs/1509.00116>`_ (Eq. 7).
"""

import numpy as np
from numpy.linalg import multi_dot

try:
    import torch

    torch_available = True
except ImportError:
    torch_available = False


class CodedApertureReconstruction:
    """
    Reconstruction method for the (non-iterative) Tikhonov algorithm presented in the `FlatCam paper <https://arxiv.org/abs/1509.00116>`_.

    TODO: operations in float32
    """

    def __init__(self, mask, image_shape, P=None, Q=None, lmbd=3e-4):
        """
        Parameters
        ----------
        mask : py:class:`lensless.hardware.mask.CodedAperture`
            Coded aperture mask object.
        image_shape : (`array-like` or `tuple`)
            The shape of the image to reconstruct.
        P : :py:class:`~numpy.ndarray`, optional
            Left convolution matrix in measurement operator. Must be of shape (measurement_resolution[0], image_shape[0]).
            By default, it is generated from the mask. In practice, it may be useful to measure as in the FlatCam paper.
        Q : :py:class:`~numpy.ndarray`, optional
            Right convolution matrix in measurement operator. Must be of shape (measurement_resolution[1], image_shape[1]).
            By default, it is generated from the mask. In practice, it may be useful to measure as in the FlatCam paper.
        lmbd: float:
            Regularization parameter. Default value is `3e-4` as in the FlatCam paper `code <https://github.com/tanjasper/flatcam/blob/master/python/demo.py>`_.
        """

        self.lmbd = lmbd
        if P is None or Q is None:
            self.P, self.Q = mask.get_conv_matrices(image_shape)
        else:
            self.P = P
            self.Q = Q
        assert self.P.shape == (
            mask.resolution[0],
            image_shape[0],
        ), "Left matrix P shape mismatch"
        assert self.Q.shape == (
            mask.resolution[1],
            image_shape[1],
        ), "Right matrix Q shape mismatch"

    def apply(self, img):
        """
        Method for performing Tikhinov reconstruction.

        Parameters
        ----------
        img : :py:class:`~numpy.ndarray` or :py:class:`torch.Tensor`
            Lensless capture measurement. Must be 3D even if grayscale.

        Returns
        -------
        :py:class:`~numpy.ndarray` or :py:class:`~torch.Tensor`
            Reconstructed image, in the same format as the measurement.
        """
        assert (
            len(img.shape) == 3
        ), "Object should be a 3D array or tensor (HxWxC) even if grayscale."

        if torch_available and isinstance(img, torch.Tensor):

            # Empty matrix for reconstruction
            n_channels = img.shape[-1]
            x_est = torch.empty([self.P.shape[1], self.Q.shape[1], n_channels])

            self.P = torch.from_numpy(self.P).float()
            self.Q = torch.from_numpy(self.Q).float()

            # Applying reconstruction for each channel
            for c in range(n_channels):
                Yc = img[:, :, c]

                # SVD of left matrix
                UL, SL, VLh = torch.linalg.svd(self.P)
                VL = VLh.T
                DL = torch.cat(
                    (
                        torch.diag(SL),
                        torch.zeros([self.P.shape[0] - SL.size(0), SL.size(0)], device=SL.device),
                    )
                )
                singLsq = SL**2

                # SVD of right matrix
                UR, SR, VRh = torch.linalg.svd(self.Q)
                VR = VRh.T
                DR = torch.cat(
                    (
                        torch.diag(SR),
                        torch.zeros([self.Q.shape[0] - SR.size(0), SR.size(0)], device=SR.device),
                    )
                )
                singRsq = SR**2

                # Applying analytical reconstruction
                inner = torch.linalg.multi_dot([DL.T, UL.T, Yc, UR, DR]) / (
                    torch.outer(singLsq, singRsq) + torch.full(x_est.shape[0:2], self.lmbd)
                )
                x_est[:, :, c] = torch.linalg.multi_dot([VL, inner, VR.T])

            # Non-negativity constraint: setting all negative values to 0
            x_est = torch.clamp(x_est, min=0)

            # Normalizing the image
            x_est = (x_est - x_est.min()) / (x_est.max() - x_est.min())

        else:

            # Empty matrix for reconstruction
            n_channels = img.shape[-1]
            x_est = np.empty([self.P.shape[1], self.Q.shape[1], n_channels])

            # Applying reconstruction for each channel
            for c in range(n_channels):

                # SVD of left matrix
                UL, SL, VLh = np.linalg.svd(self.P, full_matrices=True)
                VL = VLh.T
                DL = np.concatenate((np.diag(SL), np.zeros([self.P.shape[0] - SL.size, SL.size])))
                singLsq = np.square(SL)

                # SVD of right matrix
                UR, SR, VRh = np.linalg.svd(self.Q, full_matrices=True)
                VR = VRh.T
                DR = np.concatenate((np.diag(SR), np.zeros([self.Q.shape[0] - SR.size, SR.size])))
                singRsq = np.square(SR)

                # Applying analytical reconstruction
                Yc = img[:, :, c]
                inner = multi_dot([DL.T, UL.T, Yc, UR, DR]) / (
                    np.outer(singLsq, singRsq) + np.full(x_est.shape[0:2], self.lmbd)
                )
                x_est[:, :, c] = multi_dot([VL, inner, VR.T])

            # Non-negativity constraint: setting all negative values to 0
            x_est = x_est.clip(min=0)

            # Normalizing the image
            x_est = (x_est - x_est.min()) / (x_est.max() - x_est.min())

        return x_est



"""
DigiCam example to remotely:
1. Set mask pattern.
2. Capture image.
3. Reconstruct image with simulated PSF.

TODO: display image. At the moment should be done with `scripts/measure/remote_display.py`

"""


import hydra
from hydra.utils import to_absolute_path
import numpy as np
from lensless.hardware.slm import set_programmable_mask, adafruit_sub2full
from lensless.hardware.utils import capture
import torch
from lensless import ADMM
from lensless.utils.io import save_image
from lensless.hardware.trainable_mask import AdafruitLCD
from lensless.utils.io import load_image, load_psf
from lensless.utils.image import gamma_correction


@hydra.main(version_base=None, config_path="../../configs", config_name="digicam_example")
def digicam(config):
    measurement_fp = config.capture.fp
    psf_fp = config.psf
    mask_fp = config.mask.fp
    seed = config.mask.seed
    rpi_username = config.rpi.username
    rpi_hostname = config.rpi.hostname
    mask_shape = config.mask.shape
    mask_center = config.mask.center
    torch_device = config.recon.torch_device
    capture_config = config.capture
    simulation_config = config.simulation

    # load mask
    if mask_fp is not None:
        mask_vals = np.load(to_absolute_path(mask_fp))
    else:
        # create random mask within [0, 1]
        np.random.seed(seed)
        mask_vals = np.random.uniform(0, 1, mask_shape)

    # create mask
    mask = AdafruitLCD(
        initial_vals=torch.from_numpy(mask_vals.astype(np.float32)),
        sensor=capture_config["sensor"],
        slm="adafruit",
        downsample=capture_config["down"],
        flipud=capture_config["flip"],
        use_waveprop=simulation_config.get("use_waveprop", False),
        scene2mask=simulation_config.get("scene2mask", None),
        mask2sensor=simulation_config.get("mask2sensor", None),
        deadspace=simulation_config.get("deadspace", True),
        # color_filter=color_filter,
    )

    # use measured PSF or simulate
    if psf_fp is not None:
        psf = load_psf(
            fp=to_absolute_path(psf_fp),
            downsample=capture_config["down"],
            flip=capture_config["flip"],
        )
        psf = torch.from_numpy(psf).type(torch.float32).to(torch_device)
    else:
        psf = mask.get_psf().to(torch_device).detach()
    psf_np = psf[0].cpu().numpy()
    psf_fp = "digicam_psf.png"

    gamma = simulation_config.get("gamma", None)
    if gamma is not None:
        psf_np = psf_np / psf_np.max()
        psf_np = gamma_correction(psf_np, gamma=gamma)

    save_image(psf_np, psf_fp)
    print(f"PSF shape: {psf.shape}")
    print(f"PSF saved to {psf_fp}")

    if measurement_fp is not None:
        # load image
        img = load_image(
            to_absolute_path(measurement_fp),
            verbose=True,
        )

    else:
        ## measure data
        # -- prepare full mask
        pattern = adafruit_sub2full(
            mask_vals,
            center=mask_center,
        )

        # -- set mask
        print("Setting mask...")
        set_programmable_mask(
            pattern,
            "adafruit",
            rpi_username=rpi_username,
            rpi_hostname=rpi_hostname,
        )

        # -- capture
        print("Capturing...")
        localfile, img = capture(
            rpi_username=rpi_username,
            rpi_hostname=rpi_hostname,
            verbose=False,
            **capture_config,
        )
        print(f"Captured to {localfile}")

    """ analyze image """
    print("image range: ", img.min(), img.max())

    """ reconstruction """
    # -- normalize
    img = img.astype(np.float32) / img.max()
    # prep
    img = torch.from_numpy(img)
    # -- if [H, W, C] -> [D, H, W, C]
    if len(img.shape) == 3:
        img = img.unsqueeze(0)
    if capture_config["flip"]:
        img = torch.rot90(img, dims=(-3, -2), k=2)

    # reconstruct
    print("Reconstructing")
    recon = ADMM(psf)
    recon.set_data(img.to(psf.device))
    res = recon.apply(disp_iter=None, plot=False, n_iter=config.recon.n_iter)
    res_np = res[0].cpu().numpy()
    res_np = res_np / res_np.max()
    lensless_np = img[0].cpu().numpy()
    save_image(lensless_np, "digicam_raw.png")
    save_image(res_np, "digicam_recon.png")

    print("Done")


if __name__ == "__main__":
    digicam()



"""
Prep image for displaying in full screen for DiffuserCam capture.

```
python scripts/measure/prep_display_image.py --fp data/original_images/rect.jpg \
--pad 50 --output_path test.jpg
```

"""

import cv2
import numpy as np
from PIL import Image
import click


@click.command()
@click.option(
    "--fp",
    type=str,
    help="Path of file to display.",
)
@click.option(
    "--pad",
    default=0,
    type=float,
    help="Padding percentage along each dimension.",
)
@click.option(
    "--output_path",
    default=None,
)
@click.option(
    "--vshift",
    default=0,
    type=float,
    help="Vertical shift percentage.",
)
@click.option(
    "--hshift",
    default=0,
    type=float,
    help="Horizontal shift percentage.",
)
@click.option(
    "--brightness",
    default=100,
    type=float,
    help="Brightness percentage.",
)
@click.option(
    "--screen_res",
    default=None,
    nargs=2,
    type=int,
    help="Screen resolution in pixels (width, height).",
)
@click.option(
    "--rot90",
    default=0,
    type=int,
    help="How many times to rotate provided image by 90 degrees.",
)
@click.option(
    "--landscape",
    is_flag=True,
    help="Force landscape.",
)
@click.option(
    "--image_res",
    default=None,
    nargs=2,
    type=int,
    help="Image resolution in pixels (width, height).",
)
def display(
    fp, pad, output_path, vshift, brightness, screen_res, hshift, rot90, landscape, image_res
):
    interpolation = cv2.INTER_NEAREST

    # load image
    img_og = cv2.imread(fp, cv2.IMREAD_UNCHANGED)
    img_og = cv2.cvtColor(img_og, cv2.COLOR_BGR2RGB)
    if landscape:
        if img_og.shape[0] > img_og.shape[1]:
            img_og = np.rot90(img_og)

    # rotate image
    if rot90:
        img_og = np.rot90(img_og, k=rot90)

        # if odd, swap hshift and vshift
        if rot90 % 2:
            vshift, hshift = hshift, vshift

    if screen_res:
        image_height, image_width = img_og.shape[:2]
        img = np.zeros((screen_res[1], screen_res[0], 3), dtype=img_og.dtype)

        if image_res is None:

            # set image with padding and correct aspect ratio
            if screen_res[0] < screen_res[1]:

                max_ratio = screen_res[1] / float(image_height)

                new_width = int(screen_res[0] / (1 + 2 * pad / 100))
                ratio = new_width / float(image_width)
                # new_height = int(ratio * image_height)

            else:

                max_ratio = screen_res[0] / float(image_width)

                new_height = int(screen_res[1] / (1 + 2 * pad / 100))
                ratio = new_height / float(image_height)
                # new_width = int(ratio * image_width)

            ratio = min(ratio, max_ratio)
            new_width = int(ratio * image_width)
            new_height = int(ratio * image_height)
            image_res = (new_width, new_height)

        # if negative value in image res
        elif image_res[0] < 0 or image_res[1] < 0:
            assert image_res[0] > 0 or image_res[1] > 0, "Both dimensions cannot be negative."
            # rescale according to non-negative value
            if image_res[0] < 0:
                new_height = image_res[1]
                ratio = new_height / float(image_height)
                image_res = (int(ratio * image_width), new_height)

            elif image_res[1] < 0:
                new_width = image_res[0]
                ratio = new_width / float(image_width)
                image_res = (new_width, int(ratio * image_height))

        # set image within screen
        img_og = cv2.resize(img_og, image_res, interpolation=interpolation)
        img[: image_res[1], : image_res[0]] = img_og

        # center
        img = np.roll(img, shift=int((screen_res[1] - image_res[1]) / 2), axis=0)
        img = np.roll(img, shift=int((screen_res[0] - image_res[0]) / 2), axis=1)

    else:
        # pad image
        if pad:
            padding_amount = np.array(img_og.shape[:2]) * pad / 100
            pad_width = (
                (int(padding_amount[0] // 2), int(padding_amount[0] // 2)),
                (int(padding_amount[1] // 2), int(padding_amount[1] // 2)),
                (0, 0),
            )
            img = np.pad(img_og, pad_width=pad_width)
        else:
            img = img_og

    if vshift:
        nx, _, _ = img.shape
        img = np.roll(img, shift=int(vshift * nx / 100), axis=0)

    if hshift:
        _, ny, _ = img.shape
        img = np.roll(img, shift=int(hshift * ny / 100), axis=1)

    if brightness:
        img = (img * brightness / 100).astype(np.uint8)

    # save to file
    im = Image.fromarray(img)
    im.save(output_path)


if __name__ == "__main__":
    display()



"""Modified from https://github.com/Waller-Lab/DiffuserCam-Tutorial/blob/master/ADMM.py"""

import numpy as np
import matplotlib.pyplot as plt
import pathlib as plib
from lensless.utils.plot import plot_image


def U_update(eta, image_est, param):
    """Total variation update."""
    return soft_thresh(Psi(image_est) + eta / param["mu2"], param["tau"] / param["mu2"])


def soft_thresh(x, thresh):
    # numpy automatically applies functions to each element of the array
    return np.sign(x) * np.maximum(0, np.abs(x) - thresh)


def Psi(v):
    """Gradient of image estimate, approximated by finite difference."""
    return np.stack((np.roll(v, 1, axis=0) - v, np.roll(v, 1, axis=1) - v), axis=2)


def PsiT(U):
    """Adjoint of of finite difference, corresponds to sum of two background differences."""
    diff1 = np.roll(U[..., 0], -1, axis=0) - U[..., 0]
    diff2 = np.roll(U[..., 1], -1, axis=1) - U[..., 1]
    return diff1 + diff2


def X_update(xi, image_est, H_fft, sensor_reading, X_divmat, param):
    return X_divmat * (xi + param["mu1"] * conv(image_est, H_fft) + crop_adj(sensor_reading, param))


def conv(vk, H_fft):
    """Convolution operator."""
    return np.real(np.fft.fftshift(np.fft.ifft2(np.fft.fft2(np.fft.ifftshift(vk)) * H_fft)))


def conv_adj(x, H_fft):
    """Adjoint of convolution operator."""
    x_zeroed = np.fft.ifftshift(x)
    return np.real(np.fft.fftshift(np.fft.ifft2(np.fft.fft2(x_zeroed) * np.conj(H_fft))))


def crop(sensor_reading, param):
    """Crop operation."""
    # Image stored as matrix (row-column rather than x-y)
    top = (param["full_size"][0] - param["sensor_size"][0]) // 2
    bottom = (param["full_size"][0] + param["sensor_size"][0]) // 2
    left = (param["full_size"][1] - param["sensor_size"][1]) // 2
    right = (param["full_size"][1] + param["sensor_size"][1]) // 2
    return sensor_reading[top:bottom, left:right]


def crop_adj(sensor_reading, param):
    """Adjoint of crop, namely zero-padding."""

    v_pad = (param["full_size"][0] - param["sensor_size"][0]) // 2
    h_pad = (param["full_size"][1] - param["sensor_size"][1]) // 2
    return np.pad(
        sensor_reading, ((v_pad, v_pad), (h_pad, h_pad)), "constant", constant_values=(0, 0)
    )


def precompute_X_divmat(param):
    """Only call this function once!
    Store it in a variable and only use that variable
    during every update step"""
    return 1.0 / (crop_adj(np.ones(param["sensor_size"]), param) + param["mu1"])


def W_update(rho, image_est, param):
    """Non-negativity update"""
    return np.maximum(rho / param["mu3"] + image_est, 0)


def r_calc(w, rho, u, eta, x, xi, H_fft, param):
    return (
        (param["mu3"] * w - rho)
        + PsiT(param["mu2"] * u - eta)
        + conv_adj(param["mu1"] * x - xi, H_fft)
    )


def V_update(w, rho, u, eta, x, xi, H_fft, R_divmat, param):
    """Image update estimate."""
    freq_space_result = R_divmat * np.fft.fft2(
        np.fft.ifftshift(r_calc(w, rho, u, eta, x, xi, H_fft, param))
    )
    return np.real(np.fft.fftshift(np.fft.ifft2(freq_space_result)))


def precompute_PsiTPsi(param):
    PsiTPsi = np.zeros(param["full_size"])
    PsiTPsi[0, 0] = 4
    PsiTPsi[0, 1] = PsiTPsi[1, 0] = PsiTPsi[0, -1] = PsiTPsi[-1, 0] = -1
    PsiTPsi = np.fft.fft2(PsiTPsi)
    return PsiTPsi


def precompute_R_divmat(H_fft, PsiTPsi, param):
    """Only call this function once!
    Store it in a variable and only use that variable
    during every update step"""
    MTM_component = param["mu1"] * (np.abs(np.conj(H_fft) * H_fft))
    PsiTPsi_component = param["mu2"] * np.abs(PsiTPsi)
    id_component = param["mu3"]
    """This matrix is a mask in frequency space. So we will only use
    it on images that have already been transformed via an fft"""
    return 1.0 / (MTM_component + PsiTPsi_component + id_component)


def xi_update(xi, V, H_fft, X, param):
    return xi + param["mu1"] * (conv(V, H_fft) - X)


def eta_update(eta, V, U, param):
    return eta + param["mu2"] * (Psi(V) - U)


def rho_update(rho, V, W, param):
    return rho + param["mu3"] * (V - W)


def init_matrices(H_fft, param):
    X = np.zeros(param["full_size"])
    U = np.zeros((param["full_size"][0], param["full_size"][1], 2))
    V = np.zeros(param["full_size"])
    W = np.zeros(param["full_size"])

    xi = np.zeros_like(conv(V, H_fft))
    eta = np.zeros_like(Psi(V))
    rho = np.zeros_like(W)
    return X, U, V, W, xi, eta, rho


def precompute_H_fft(psf, param):
    return np.fft.fft2(np.fft.ifftshift(crop_adj(psf, param)))


def admm_step(X, U, V, W, xi, eta, rho, precomputed, param):
    H_fft, data, X_divmat, R_divmat = precomputed
    U = U_update(eta, V, param)  # TODO : pass U or V?
    X = X_update(xi, V, H_fft, data, X_divmat, param)  # TODO : pass X or V?
    V = V_update(W, rho, U, eta, X, xi, H_fft, R_divmat, param)
    W = W_update(rho, V, param)
    xi = xi_update(xi, V, H_fft, X, param)
    eta = eta_update(eta, V, U, param)
    rho = rho_update(rho, V, W, param)

    return X, U, V, W, xi, eta, rho


def apply_admm(psf, data, param, n_iter=5, disp_iter=1, plot_pause=0.5, plot=True, save=False):
    """
    ADMM reconstruction approach.

    Parameters
    ----------
    psf :py:class:`~numpy.ndarray`
        2-D array of PSF.
    data :py:class:`~numpy.ndarray`
        2-D array of raw measurement data.
    param : dict
        ADMM parameters
    n_iter : int, optional
        Number of iterations.
    disp_iter : int, optional
        How many iterations to wait before plotting. Set to None for no intermediate plots.
    plot_pause : float, optional
        How much to pause (in seconds) between iterations plot.

    Returns
    -------
    final_im :py:class:`~numpy.ndarray`
        2-D array of reconstructed data.
    ax : :py:class:`~matplotlib.axes.Axes`
    """
    assert "mu1" in param.keys()
    assert "mu2" in param.keys()
    assert "mu3" in param.keys()
    assert "tau" in param.keys()
    assert "sensor_size" in param.keys()
    assert "full_size" in param.keys()

    assert len(psf.shape) == len(data.shape)
    is_rgb = True if len(psf.shape) == 3 else False
    if not is_rgb:
        psf = psf[:, :, np.newaxis]
        data = data[:, :, np.newaxis]

    # precompute some matrices
    H_fft = []
    X = []
    U = []
    V = []
    W = []
    xi = []
    eta = []
    rho = []
    X_divmat = []
    PsiTPsi = []
    R_divmat = []
    for c in range(psf.shape[2]):
        H_fft.append(precompute_H_fft(psf[:, :, c], param))
        _X, _U, _V, _W, _xi, _eta, _rho = init_matrices(H_fft[c], param)
        X.append(_X)
        U.append(_U)
        V.append(_V)
        W.append(_W)
        xi.append(_xi)
        eta.append(_eta)
        rho.append(_rho)
        X_divmat.append(precompute_X_divmat(param))
        PsiTPsi.append(precompute_PsiTPsi(param))
        R_divmat.append(precompute_R_divmat(H_fft[c], PsiTPsi[c], param))

    if plot and disp_iter is not None:
        if is_rgb:
            ax = plot_image(data)
        else:
            ax = plot_image(data[:, :, 0])
    else:
        ax = None
        disp_iter = n_iter + 1

    for i in range(n_iter):
        for c in range(psf.shape[2]):
            X[c], U[c], V[c], W[c], xi[c], eta[c], rho[c] = admm_step(
                X[c],
                U[c],
                V[c],
                W[c],
                xi[c],
                eta[c],
                rho[c],
                precomputed=[H_fft[c], data[:, :, c], X_divmat[c], R_divmat[c]],
                param=param,
            )

        if plot and (i + 1) % disp_iter == 0:
            if is_rgb:
                image = np.zeros(tuple(param["sensor_size"]) + (3,))
                for c in range(3):
                    image[:, :, c] = crop(V[c], param)
            else:
                image = crop(V[0], param)
            image[image < 0] = 0
            ax = plot_image(image, ax=ax)
            ax.set_title("Reconstruction after iteration {}".format(i + 1))
            if save:
                plt.savefig(plib.Path(save) / f"{i + 1}_diffusercam.png")
            plt.draw()
            plt.pause(plot_pause)

    if is_rgb:
        final_im = np.zeros(tuple(param["sensor_size"]) + (3,))
        for c in range(3):
            final_im[:, :, c] = crop(V[c], param)
    else:
        final_im = crop(V[0], param)
    final_im[final_im < 0] = 0
    if plot:
        ax = plot_image(final_im, ax=ax)
        ax.set_title("Final reconstruction after {} iterations".format(n_iter))
        if save:
            plt.savefig(plib.Path(save) / f"{n_iter}_diffusercam.png")
        return final_im, ax
    else:
        return final_im
